import datetime
import io
import json
import typing as t
from contextlib import suppress

from pulp_glue.common.i18n import get_translation

translation = get_translation(__package__)
_ = translation.gettext

ISO_DATE_FORMAT = "%Y-%m-%d"
ISO_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class SchemaError(ValueError):
    pass


class ValidationError(ValueError):
    pass


class OpenApi3JsonEncoder(json.JSONEncoder):
    def default(self, o: t.Any) -> t.Any:
        if isinstance(o, datetime.datetime):
            return o.strftime(ISO_DATETIME_FORMAT)
        elif isinstance(o, datetime.date):
            return o.strftime(ISO_DATE_FORMAT)
        else:
            return super().default(o)


def encode_json(o: t.Any) -> str:
    return json.dumps(o, cls=OpenApi3JsonEncoder)


def encode_param(value: t.Any) -> t.Any:
    if isinstance(value, list):
        return [encode_param(item) for item in value]
    if isinstance(value, datetime.datetime):
        return value.strftime(ISO_DATETIME_FORMAT)
    elif isinstance(value, datetime.date):
        return value.strftime(ISO_DATE_FORMAT)
    else:
        return value


def _assert_type(
    name: str,
    value: t.Any,
    types: t.Union[t.Type[object], t.Tuple[t.Type[object], ...]],
    type_name: str,
) -> None:
    if not isinstance(value, types):
        raise ValidationError(
            _("'{name}' is expected to be a {type_name}.").format(name=name, type_name=type_name)
        )


def _assert_min_max(schema: t.Any, name: str, value: t.Any) -> None:
    if (minimum := schema.get("minimum")) is not None:
        if schema.get("exclusiveMinimum", False):
            if minimum >= value:
                raise ValidationError(
                    _("'{name}' is expected to be larger than {minimum}").format(
                        name=name, minimum=minimum
                    )
                )
        else:
            if minimum > value:
                raise ValidationError(
                    _("'{name}' is expected to not be smaller than {minimum}").format(
                        name=name, minimum=minimum
                    )
                )
    if (maximum := schema.get("maximum")) is not None:
        if schema.get("exclusiveMaximum", False):
            if maximum <= value:
                raise ValidationError(
                    _("'{name}' is expected to be smaller than {maximum}").format(
                        name=name, maximum=maximum
                    )
                )
        else:
            if maximum < value:
                raise ValidationError(
                    _("'{name}' is expected to not be larger than {maximum}").format(
                        name=name, maximum=maximum
                    )
                )


def validate(schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]) -> None:
    if (schema_ref := schema.get("$ref")) is not None:
        # From json-schema:
        # "All other properties in a "$ref" object MUST be ignored."
        _validate_ref(schema_ref, name, value, components)
        return

    if value is None:
        # This seems to be the openapi 3.0.3 way.
        # in 3.1.* they use `"type": ["string", "null"]` instead.
        if schema.get("nullable", False):
            return

    if (schema_type := schema.get("type")) is not None:
        if isinstance(schema_type, list):
            if len(schema_type) == 0:
                raise SchemaError(_("{name} specified an empty type array").format(name=name))
            errors = []
            for stype in schema_type:
                try:
                    _validate_type(stype, schema, name, value, components)
                    break
                except ValidationError as e:
                    errors.append(f"{stype}: {e}")
            else:
                raise ValidationError(
                    _("{name} did not match any of the types: {errors}").format(
                        name=name, errors="\n".join(errors)
                    )
                )

        else:
            _validate_type(schema_type, schema, name, value, components)

    # allOf etc allow for composition, but the spec isn't particularly clear about that.
    if (all_of := schema.get("allOf")) is not None:
        if isinstance(value, dict):
            value = value.copy()
            for sub_schema in all_of:
                validate(sub_schema, name, value, components)
        else:
            # This may not even be a valid case according to the specification.
            # But it is used by drf-spectacular to reference enum-strings.
            for sub_schema in all_of:
                validate(sub_schema, name, value, components)

    if (any_of := schema.get("anyOf")) is not None:
        for sub_schema in any_of:
            with suppress(ValidationError):
                validate(sub_schema, name, value, components)
                break
        else:
            raise ValidationError(
                _("'{name}' does not match any of the provide schemata.").format(name=name)
            )


def _validate_type(
    schema_type: str, schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    if (typed_validator := _TYPED_VALIDATORS.get(schema_type)) is not None:
        typed_validator(schema, name, value, components)
    else:
        raise NotImplementedError(
            _("Type `{schema_type}` is not implemented yet.").format(schema_type=schema_type)
        )


def _validate_ref(schema_ref: str, name: str, value: t.Any, components: t.Dict[str, t.Any]) -> None:
    if not schema_ref.startswith("#/components/schemas/"):
        raise SchemaError(_("'{name}' contains an invalid reference.").format(name=name))
    schema_name = schema_ref[21:]
    validate(components[schema_name], name, value, components)


def _validate_array(schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]) -> None:
    _assert_type(name, value, list, "array")
    if (min_items := schema.get("minItems")) is not None:
        if len(value) < min_items:
            raise ValidationError(
                _("'{name}' is expected to have at least {min_items} items.").format(
                    name=name, min_items=min_items
                )
            )
    if (max_items := schema.get("maxItems")) is not None:
        if len(value) > max_items:
            raise ValidationError(
                _("'{name}' is expected to have at most {max_items} items.").format(
                    name=name, max_items=max_items
                )
            )
    if schema.get("uniqueItems", False):
        if len(set(value)) != len(value):
            raise ValidationError(_("'{name}' is expected to have unique items.").format(name=name))

    for i, item in enumerate(value):
        validate(schema["items"], f"{name}[{i}]", item, components)


def _validate_boolean(
    schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    _assert_type(name, value, bool, "boolean")


def _validate_integer(
    schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    _assert_type(name, value, int, "integer")
    _assert_min_max(schema, name, value)

    if (multiple_of := schema.get("multipleOf")) is not None:
        if value % multiple_of != 0:
            raise ValidationError(
                _("'{name}' is expected to be a multiple of {multiple_of}").format(
                    name=name, multiple_of=multiple_of
                )
            )


def _validate_null(schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]) -> None:
    if value is not None:
        raise ValidationError(_("'{name}' is expected to be a null").format(name=name))


def _validate_number(
    schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    _assert_type(name, value, (float, int), "number")
    _assert_min_max(schema, name, value)


def _validate_object(
    schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    _assert_type(name, value, dict, "object")
    extra_values = {}
    properties = schema.get("properties", {})
    for pname, pvalue in value.items():
        if (pschema := properties.get(pname)) is not None:
            validate(pschema, f"{name}[{pname}]", pvalue, components)
        else:
            extra_values[pname] = pvalue
    additional_properties: t.Union[bool, t.Dict[str, t.Any]] = schema.get(
        "additionalProperties", {}
    )
    if len(extra_values) > 0:
        if additional_properties is False:
            raise ValidationError(
                _("'{name}' does not allow additional properties.").format(name=name)
            )
        elif isinstance(additional_properties, dict):
            for pname, pvalue in extra_values.items():
                validate(additional_properties, f"{name}[{pname}]", pvalue, components)
    if (required := schema.get("required")) is not None:
        if missing_keys := set(required) - set(value.keys()):
            raise ValidationError(
                _("'{name}' is missing properties ({missing}).").format(
                    name=name, missing=", ".join(missing_keys)
                )
            )


def _validate_string(
    schema: t.Any, name: str, value: t.Any, components: t.Dict[str, t.Any]
) -> None:
    schema_format = schema.get("format")
    if schema_format == "byte":
        _assert_type(name, value, bytes, "bytes")
    elif schema_format == "binary":
        # This is not really useful for json serialization.
        # It is there for file transfer, e.g. in multipart.
        _assert_type(name, value, (bytes, io.BufferedReader, io.BytesIO), "binary")
    elif schema_format == "date":
        _assert_type(name, value, datetime.date, "date")
    elif schema_format == "date-time":
        _assert_type(name, value, datetime.datetime, "date-time")
    else:
        _assert_type(name, value, str, "string")
    if (enum := schema.get("enum")) is not None:
        if value not in enum:
            raise ValidationError(
                _("'{name}' is expected to be one of [{enums}].").format(
                    name=name, enums=", ".join(enum)
                )
            )


_TYPED_VALIDATORS = {
    "array": _validate_array,
    "boolean": _validate_boolean,
    "integer": _validate_integer,
    "null": _validate_null,
    "number": _validate_number,
    "object": _validate_object,
    "string": _validate_string,
}



import typing as t

from pulp_glue.common.i18n import get_translation

translation = get_translation(__package__)
_ = translation.gettext


class PulpException(Exception):
    """The base exception `pulp-glue` will emit on expected error paths."""


class PulpEntityNotFound(PulpException):
    """Exception to signify that an entity was not found."""


class PulpHTTPError(PulpException):
    """Exception to indicate HTTP error responses."""

    def __init__(self, msg: str, status_code: int, operation_id: t.Optional[str] = None) -> None:
        super().__init__(msg)
        self.status_code = status_code
        self.operation_id = operation_id


class PulpAuthenticationFailed(PulpHTTPError):
    """Exception to signal a failed authentication."""

    def __init__(self, operation_id: str) -> None:
        super().__init__(
            _("Authentication failed for {operation_id}.").format(operation_id=operation_id),
            401,
            operation_id,
        )


class PulpNotAutorized(PulpHTTPError):
    """Exception to signal an action was not autorized."""

    def __init__(self, operation_id: str) -> None:
        super().__init__(
            _("Operation {operation_id} is not authorized.").format(operation_id=operation_id),
            403,
            operation_id,
        )


class PulpNoWait(Exception):
    """Exception to indicate that a task continues running in the background."""


class NotImplementedFake(NotImplementedError, PulpException):
    """Exception to indicat that a call without a fake version was attempted in fake_mode."""


class OpenAPIError(PulpException):
    """Base Exception for errors related to using the openapi spec."""


class ValidationError(OpenAPIError):
    """Exception raised for failed client side validation of parameters or request bodies."""


class UnsafeCallError(OpenAPIError):
    """Exception raised for POST, PUT, PATCH or DELETE calls with `safe_calls_only=True`."""



import os
import typing as t

from pulp_glue.common.context import (
    BATCH_SIZE,
    EntityDefinition,
    PluginRequirement,
    PulpACSContext,
    PulpContentContext,
    PulpDistributionContext,
    PulpEntityContext,
    PulpPublicationContext,
    PulpRemoteContext,
    PulpRepositoryContext,
    PulpRepositoryVersionContext,
    PulpViewSetContext,
    api_quirk,
)
from pulp_glue.common.exceptions import PulpException
from pulp_glue.common.i18n import get_translation
from pulp_glue.common.openapi import OpenAPI

translation = get_translation(__package__)
_ = translation.gettext


@api_quirk(PluginRequirement("rpm", specifier=">=3.3.0"))
def patch_rpm_copy_scheme(api: OpenAPI) -> None:
    path, operation = next(
        (
            (k, v["post"])
            for k, v in api.api_spec["paths"].items()
            if k.endswith("/api/v3/rpm/copy/")
        )
    )
    api.api_spec["components"]["schemas"]["RpmCopy"] = {
        "type": "object",
        "description": "A serializer for Content Copy API.",
        "properties": {
            "config": {
                "description": "A JSON document describing sources, destinations,"
                " and content to be copied"
            },
            "dependency_solving": {"type": "boolean"},
        },
        "required": ["config"],
    }
    operation["operationId"] = "rpm_copy_content"
    for item in operation["requestBody"]["content"].values():
        item["schema"]["$ref"] = "#/components/schemas/RpmCopy"

    api.operations["rpm_copy_content"] = ("post", path)


class PulpRpmACSContext(PulpACSContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "rpm"
    ENTITY = _("rpm ACS")
    ENTITIES = _("rpm ACSes")
    HREF = "rpm_rpm_alternate_content_source_href"
    ID_PREFIX = "acs_rpm_rpm"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.18.0")]
    CAPABILITIES = {"roles": [PluginRequirement("rpm", specifier=">=3.19.0")]}


class PulpRpmCompsXmlContext(PulpEntityContext):
    UPLOAD_COMPS_ID: t.ClassVar[str] = "rpm_comps_upload"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.17.0")]

    def upload_comps(
        self, file: t.IO[bytes], repo_href: t.Optional[str], replace: t.Optional[bool]
    ) -> t.Any:
        self.pulp_ctx.echo(_("Uploading file {filename}").format(filename=file.name), err=True)
        file.seek(0)
        return self.call(
            "upload_comps",
            body={"repository": repo_href, "replace": replace, "file": file},
        )


class PulpRpmDistributionContext(PulpDistributionContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "rpm"
    ENTITY = _("rpm distribution")
    ENTITIES = _("rpm distributions")
    HREF = "rpm_rpm_distribution_href"
    ID_PREFIX = "distributions_rpm_rpm"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]
    CAPABILITIES = {"roles": [PluginRequirement("rpm", specifier=">=3.19.0")]}

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.16.0")):
            if "repository" in body and "publication" not in body:
                body["publication"] = None
            if "repository" not in body and "publication" in body:
                body["repository"] = None
        if body.get("generate_repo_config") is False:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.23.0",
                    feature=_("configuring the generation of the config.repo file"),
                )
            )
        elif self.pulp_ctx.has_plugin(PluginRequirement("rpm", specifier="<3.23.0")):
            body.pop("generate_repo_config", None)
        return body


class PulpRpmPackageContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "package"
    ENTITY = "rpm package"
    ENTITIES = "rpm packages"
    HREF = "rpm_package_href"
    ID_PREFIX = "content_rpm_packages"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]
    CAPABILITIES = {"upload": []}

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if partial is False:
            if body.get("relative_path") is None:
                self.pulp_ctx.needs_plugin(PluginRequirement("rpm", specifier=">=3.18.0"))
            else:
                PulpException(_("--relative-path must be provided"))
        return body

    def list_iterator(
        self,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        offset: int = 0,
        batch_size: int = BATCH_SIZE,
        stats: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Iterator[t.Any]:
        contains_startswith = [
            "name__contains",
            "name__startswith",
            "release__contains",
            "release__startswith",
            "arch__contains",
            "arch__startswith",
        ]
        if parameters is not None and any(
            v for k, v in parameters.items() if k in contains_startswith
        ):
            self.pulp_ctx.needs_plugin(
                PluginRequirement("rpm", specifier=">=3.20.0", feature=_("substring filters"))
            )
        return super().list_iterator(
            parameters=parameters,
            offset=offset,
            batch_size=batch_size,
            stats=stats,
        )

    def upload(
        self,
        file: t.IO[bytes],
        chunk_size: int,
        repository: t.Optional[PulpRepositoryContext],
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Create an RPM package by uploading a file.

        This function is deprecated. The create call can handle the upload logic transparently.

        Parameters:
            file: A file like object that supports `os.path.getsize`.
            chunk_size: Size of the chunks to upload independently.
            repository: Repository context to add the newly created content to.
            kwargs: Extra args specific to the content type, passed to the create call.

        Returns:
            The result of the create task.
        """
        self.needs_capability("upload")
        size = os.path.getsize(file.name)
        body: t.Dict[str, t.Any] = {**kwargs}

        if not self.pulp_ctx.fake_mode:
            if chunk_size > size:
                # Small file: direct upload
                body["file"] = file
            else:
                # Large file: chunked upload
                if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.20.0")):
                    from pulp_glue.core.context import PulpUploadContext

                    upload_href = PulpUploadContext(self.pulp_ctx).upload_file(file, chunk_size)
                    body["upload"] = upload_href
                else:
                    from pulp_glue.core.context import PulpArtifactContext

                    artifact_href = PulpArtifactContext(self.pulp_ctx).upload(file, chunk_size)
                    body["artifact"] = artifact_href

        # For rpm plugin >= 3.32.5, use synchronous upload endpoint when no repository is provided
        # For older versions, always use the create endpoint (backward compatibility)
        if repository is None and self.pulp_ctx.has_plugin(
            PluginRequirement("rpm", specifier=">=3.32.5")
        ):
            return self.call("upload", body=body)

        # Repository is specified or older rpm version: use create endpoint (async path)
        if repository is not None:
            body["repository"] = repository
        return self.create(body=body)


class PulpRpmAdvisoryContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "update_record"
    ENTITY = "rpm advisory"
    ENTITIES = "rpm advisories"
    HREF = "rpm_update_record_href"
    ID_PREFIX = "content_rpm_advisories"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmDistributionTreeContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "distribution_tree"
    ENTITY = "rpm distribution tree"
    ENTITIES = "rpm distribution trees"
    HREF = "rpm_distribution_tree_href"
    ID_PREFIX = "content_rpm_distribution_trees"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmModulemdDefaultsContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "modulemd_defaults"
    ENTITY = "rpm modulemd defaults"
    ENTITIES = "rpm modulemd defaults"
    HREF = "rpm_modulemd_defaults_href"
    ID_PREFIX = "content_rpm_modulemd_defaults"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmModulemdContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "modulemd"
    ENTITY = "rpm modulemd"
    ENTITIES = "rpm modulemds"
    HREF = "rpm_modulemd_href"
    ID_PREFIX = "content_rpm_modulemds"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmPackageCategoryContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "package_category"
    ENTITY = "rpm package category"
    ENTITIES = "rpm package categories"
    HREF = "rpm_package_category_href"
    ID_PREFIX = "content_rpm_packagecategories"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmPackageEnvironmentContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "package_environment"
    ENTITY = "rpm package environment"
    ENTITIES = "rpm package environments"
    HREF = "rpm_package_environment_href"
    ID_PREFIX = "content_rpm_packageenvironments"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmPackageGroupContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "package_group"
    ENTITY = "rpm package group"
    ENTITIES = "rpm package groups"
    HREF = "rpm_package_group_href"
    ID_PREFIX = "content_rpm_packagegroups"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmPackageLangpacksContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "package_langpack"
    ENTITY = "rpm package langpack"
    ENTITIES = "rpm package langpacks"
    HREF = "rpm_package_langpacks_href"
    ID_PREFIX = "content_rpm_packagelangpacks"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmRepoMetadataFileContext(PulpContentContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "repo_metadata_file"
    ENTITY = "rpm repo metadata file"
    ENTITIES = "rpm repo metadata files"
    HREF = "rpm_repo_metadata_file_href"
    ID_PREFIX = "content_rpm_repo_metadata_files"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmPublicationContext(PulpPublicationContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "rpm"
    ENTITY = _("rpm publication")
    ENTITIES = _("rpm publications")
    HREF = "rpm_rpm_publication_href"
    ID_PREFIX = "publications_rpm_rpm"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]
    CAPABILITIES = {"roles": [PluginRequirement("rpm", specifier=">=3.19.0")]}

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        version = body.pop("version", None)
        if version is not None:
            repository_href = body.pop("repository")
            body["repository_version"] = f"{repository_href}versions/{version}/"
        if "repo_config" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.24.0",
                    feature=_("customization of the config.repo file"),
                )
            )
        if "sqlite_metadata" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.25.0",
                    inverted=True,
                    feature=_("sqlite_metadata generation"),
                )
            )
        if "metadata_checksum_type" in body or "package_checksum_type" in body:
            metadata_checksum_type = body.get("metadata_checksum_type")
            package_checksum_type = body.get("metadata_checksum_type")
            disallowed_checksums = {"md5", "sha1", "sha224"}

            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.30.0",
                    inverted=True,
                    feature=_("package_checksum_type/metadata_checksum_type"),
                )
            )

            if metadata_checksum_type and metadata_checksum_type in disallowed_checksums:
                self.pulp_ctx.needs_plugin(
                    PluginRequirement(
                        "rpm", specifier=">=3.25.0", inverted=True, feature=_("weak checksums")
                    )
                )
            if package_checksum_type and package_checksum_type in disallowed_checksums:
                self.pulp_ctx.needs_plugin(
                    PluginRequirement(
                        "rpm", specifier=">=3.25.0", inverted=True, feature=_("weak checksums")
                    )
                )
        if "repo_gpgcheck" in body or "gpgcheck" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.30.0",
                    inverted=True,
                    feature=_("gpgcheck/repo_gpgcheck"),
                )
            )

        return body


class PulpRpmRemoteContext(PulpRemoteContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "rpm"
    ENTITY = _("rpm remote")
    ENTITIES = _("rpm remotes")
    HREF = "rpm_rpm_remote_href"
    ID_PREFIX = "remotes_rpm_rpm"
    NULLABLES = PulpRemoteContext.NULLABLES | {"sles_auth_token"}
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]
    CAPABILITIES = {"roles": [PluginRequirement("rpm", specifier=">=3.19.0")]}


class PulpUlnRemoteContext(PulpRemoteContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "uln"
    ENTITY = _("uln remote")
    ENTITIES = _("uln remotes")
    HREF = "rpm_uln_remote_href"
    ID_PREFIX = "remotes_rpm_uln"
    NULLABLES = PulpRemoteContext.NULLABLES | {"uln-server-base-url"}
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.12.0")]
    CAPABILITIES = {"roles": [PluginRequirement("rpm", specifier=">=3.19.0")]}


class PulpRpmRepositoryVersionContext(PulpRepositoryVersionContext):
    HREF = "rpm_rpm_repository_version_href"
    ID_PREFIX = "repositories_rpm_rpm_versions"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]


class PulpRpmRepositoryContext(PulpRepositoryContext):
    PLUGIN = "rpm"
    RESOURCE_TYPE = "rpm"
    HREF = "rpm_rpm_repository_href"
    ID_PREFIX = "repositories_rpm_rpm"
    ENTITY = _("rpm repository")
    ENTITIES = _("rpm repositories")
    VERSION_CONTEXT = PulpRpmRepositoryVersionContext
    CAPABILITIES = {
        "sync": [],
        "pulpexport": [],
        "roles": [PluginRequirement("rpm", specifier=">=3.19.0")],
    }
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.9.0")]

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if "autopublish" in body:
            self.pulp_ctx.needs_plugin(PluginRequirement("rpm", specifier=">=3.12.0"))
        if "repo_config" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.24.0",
                    feature=_("customization of the config.repo file"),
                )
            )
        if "sqlite_metadata" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.25.0",
                    inverted=True,
                    feature=_("sqlite_metadata generation"),
                )
            )
        if "metadata_checksum_type" in body or "package_checksum_type" in body:
            metadata_checksum_type = body.get("metadata_checksum_type")
            package_checksum_type = body.get("metadata_checksum_type")
            disallowed_checksums = {"md5", "sha1", "sha224"}

            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.30.0",
                    inverted=True,
                    feature=_("package_checksum_type/metadata_checksum_type"),
                )
            )

            if metadata_checksum_type and metadata_checksum_type in disallowed_checksums:
                self.pulp_ctx.needs_plugin(
                    PluginRequirement(
                        "rpm", specifier=">=3.25.0", inverted=True, feature=_("weak checksums")
                    )
                )
            if package_checksum_type and package_checksum_type in disallowed_checksums:
                self.pulp_ctx.needs_plugin(
                    PluginRequirement(
                        "rpm", specifier=">=3.25.0", inverted=True, feature=_("weak checksums")
                    )
                )

        if "checksum_type" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.25.0",
                    feature=_("checksum_type"),
                )
            )

        if "repo_gpgcheck" in body or "gpgcheck" in body:
            self.pulp_ctx.needs_plugin(
                PluginRequirement(
                    "rpm",
                    specifier=">=3.30.0",
                    inverted=True,
                    feature=_("gpgcheck/repo_gpgcheck"),
                )
            )

        return body

    def sync(self, body: t.Optional[EntityDefinition] = None) -> t.Any:
        if body:
            if body.get("optimize") is not None:
                self.pulp_ctx.needs_plugin(PluginRequirement("rpm", specifier=">=3.3.0"))
            if body.get("sync_policy") is not None:
                self.pulp_ctx.needs_plugin(PluginRequirement("rpm", specifier=">=3.16.0"))
            if "treeinfo" in body.get("skip_types", ""):
                self.pulp_ctx.needs_plugin(
                    PluginRequirement("rpm", specifier=">=3.18.10", feature="--skip-type treeinfo")
                )

        return super().sync(body)


class PulpRpmPruneContext(PulpViewSetContext):
    ID_PREFIX: t.ClassVar[str] = "rpm_prune"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.27.0.dev")]

    def prune_packages(
        self,
        repo_hrefs: t.List[t.Union[str, PulpRpmRepositoryContext]],
        keep_days: t.Optional[int],
        dry_run: t.Optional[bool],
    ) -> t.Any:
        body: t.Dict[str, t.Any] = {
            "repo_hrefs": repo_hrefs,
            "keep_days": keep_days,
            "dry_run": dry_run,
        }
        return self.call("prune_packages", body=body)


class PulpRpmCopyContext(PulpViewSetContext):
    COPY_ID: t.ClassVar[str] = "rpm_copy_content"
    NEEDS_PLUGINS = [PluginRequirement("rpm", specifier=">=3.3.0")]

    def copy(
        self,
        config: t.List[t.Dict[str, t.Any]],
        dependency_solving: t.Optional[bool] = False,
    ) -> t.Any:
        body: t.Dict[str, t.Any] = {
            "config": config,
            "dependency_solving": dependency_solving,
        }
        return self.call("copy", body=body)



import random
import string
import typing as t

import pytest

from pulp_glue.common.context import PulpContext
from pulp_glue.common.exceptions import NotImplementedFake
from pulp_glue.file.context import PulpFileRepositoryContext

pytestmark = pytest.mark.glue


@pytest.fixture
def file_repository(pulp_ctx: PulpContext) -> t.Iterator[t.Dict[str, t.Any]]:
    name = "".join(random.choices(string.ascii_letters, k=8))
    repository_ctx = PulpFileRepositoryContext(pulp_ctx)
    yield repository_ctx.create(body={"name": name})
    repository_ctx.delete()


@pytest.mark.live
def test_fake_status(fake_pulp_ctx: PulpContext) -> None:
    fake_pulp_ctx.call("status_read")


@pytest.mark.live
def test_fake_repository_read(fake_pulp_ctx: PulpContext) -> None:
    repository_ctx = PulpFileRepositoryContext(fake_pulp_ctx)
    repository_ctx.list(limit=0, offset=0, parameters={})


@pytest.mark.live
def test_fake_repository_create(fake_pulp_ctx: PulpContext) -> None:
    name = "".join(random.choices(string.ascii_letters, k=8))
    repository_ctx = PulpFileRepositoryContext(fake_pulp_ctx)
    result = repository_ctx.create(body={"name": name})
    assert result["name"] == name
    assert result["pulp_href"] == "<FAKE ENTITY>"


@pytest.mark.live
def test_fake_repository_update(
    fake_pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any]
) -> None:
    repository_ctx = PulpFileRepositoryContext(
        fake_pulp_ctx, pulp_href=file_repository["pulp_href"]
    )
    result = repository_ctx.update(body={"description": "TEST"})
    assert result["pulp_href"] == file_repository["pulp_href"]
    assert result["name"] == file_repository["name"]
    assert result["description"] == "TEST"

    # Reload from server
    repository_ctx.pulp_href = file_repository["pulp_href"]
    repository_ctx.entity == file_repository


@pytest.mark.live
def test_fake_repository_delete(
    fake_pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any]
) -> None:
    repository_ctx = PulpFileRepositoryContext(
        fake_pulp_ctx, pulp_href=file_repository["pulp_href"]
    )
    result = repository_ctx.delete()
    assert result is None

    # Reload from server
    repository_ctx.pulp_href = file_repository["pulp_href"]
    repository_ctx.entity == file_repository


@pytest.mark.live
def test_fake_raises(fake_pulp_ctx: PulpContext) -> None:
    name = "".join(random.choices(string.ascii_letters, k=8))
    with pytest.raises(NotImplementedFake):
        fake_pulp_ctx.call("repositories_file_file_create", body={"name": name})



import typing as t
from datetime import datetime, timedelta

import requests


class OAuth2ClientCredentialsAuth(requests.auth.AuthBase):
    """
    This implements the OAuth2 ClientCredentials Grant authentication flow.
    https://datatracker.ietf.org/doc/html/rfc6749#section-4.4
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        token_url: str,
        scopes: t.Optional[t.List[str]] = None,
        verify_ssl: t.Optional[t.Union[str, bool]] = None,
    ):
        self._token_server_auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
        self._token_url = token_url
        self._scopes = scopes
        self._verify_ssl = verify_ssl

        self._access_token: t.Optional[str] = None
        self._expire_at: t.Optional[datetime] = None

    def __call__(self, request: requests.PreparedRequest) -> requests.PreparedRequest:
        if self._expire_at is None or self._expire_at < datetime.now():
            self._retrieve_token()

        assert self._access_token is not None

        request.headers["Authorization"] = f"Bearer {self._access_token}"

        # Call to untyped function "register_hook" in typed context
        request.register_hook("response", self._handle401)  # type: ignore[no-untyped-call]

        return request

    def _handle401(
        self,
        response: requests.Response,
        **kwargs: t.Any,
    ) -> requests.Response:
        if response.status_code != 401:
            return response

        # If we get this far, probably the token is not valid anymore.

        # Try to reach for a new token once.
        self._retrieve_token()

        assert self._access_token is not None

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        response.content
        response.close()
        prepared_new_request = response.request.copy()

        prepared_new_request.headers["Authorization"] = f"Bearer {self._access_token}"

        # Avoid to enter into an infinity loop.
        # Call to untyped function "deregister_hook" in typed context
        prepared_new_request.deregister_hook(  # type: ignore[no-untyped-call]
            "response", self._handle401
        )

        # "Response" has no attribute "connection"
        new_response: requests.Response = response.connection.send(prepared_new_request, **kwargs)
        new_response.history.append(response)
        new_response.request = prepared_new_request

        return new_response

    def _retrieve_token(self) -> None:
        data = {
            "grant_type": "client_credentials",
        }

        if self._scopes:
            data["scope"] = " ".join(self._scopes)

        response: requests.Response = requests.post(
            self._token_url,
            data=data,
            auth=self._token_server_auth,
            verify=self._verify_ssl,
        )

        response.raise_for_status()

        token = response.json()
        self._expire_at = datetime.now() + timedelta(seconds=token["expires_in"])
        self._access_token = token["access_token"]



import datetime
import os
import re
import sys
import time
import typing as t
import warnings
from contextlib import ExitStack

from packaging.specifiers import SpecifierSet

from pulp_glue.common.exceptions import (
    NotImplementedFake,
    OpenAPIError,
    PulpEntityNotFound,
    PulpException,
    PulpHTTPError,
    PulpNoWait,
    UnsafeCallError,
)
from pulp_glue.common.i18n import get_translation
from pulp_glue.common.openapi import BasicAuthProvider, OpenAPI

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

translation = get_translation(__package__)
_ = translation.gettext

DEFAULT_LIMIT = 25
BATCH_SIZE = 1000
DATETIME_FORMATS = [
    "%Y-%m-%dT%H:%M:%S.%fZ",  # Pulp format
    "%Y-%m-%d",  # intl. format
    "%Y-%m-%d %H:%M:%S",  # ... with time
    "%Y-%m-%dT%H:%M:%S",  # ... with time and T as a separator
    "%y/%m/%d",  # US format
    "%y/%m/%d %h:%M:%S %p",  # ... with time
    "%y/%m/%dT%h:%M:%S %p",  # ... with time and T as a separator
    "%y/%m/%d %H:%M:%S",  # ... with time 24h
    "%y/%m/%dT%H:%M:%S",  # ... with time 24h and T as a separator
    "%x",  # local format
    "%x %X",  # ... with time
    "%xT%X",  # ... with time and T as a separator
]

href_regex = re.compile(r"\/([a-z0-9-_]+\/)+", flags=re.IGNORECASE)


class PreprocessedEntityDefinition(t.Dict[str, t.Any]):
    pass


EntityDefinition = t.Union[t.Dict[str, t.Any], PreprocessedEntityDefinition]


class PluginRequirement:
    """
    A class to represent a Pulp plugin with a set of versions.

    This can be used in conjunction with `has_plugin(s)`, `needs_plugin(s)` and `CAPABILITIES`.

    Parameters:
        name: The app-label of the pluin as reported by the status API.
        feature: A string being displayed as the feature if used with `needs_plugin` and the
            condition is not met.
        inverted: Treat the version set in specifier as inverted. If no specifier is provided, this
            describes the requirement of a plugin not being installed.
        specifier: A PEP-440 compatible version range.
    """

    def __init__(
        self,
        name: str,
        feature: t.Optional[str] = None,
        inverted: bool = False,
        specifier: t.Optional[t.Union[str, SpecifierSet]] = None,
    ):
        self.name = name
        self.feature = feature
        self.inverted = inverted
        if isinstance(specifier, SpecifierSet):
            self.specifier = specifier
        else:
            self.specifier = SpecifierSet(specifier or "", prereleases=True)

    def __contains__(self, version: t.Optional[str]) -> bool:
        if version is None:
            return self.inverted
        return (version in self.specifier) != self.inverted

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name}{self.specifier})"

    __repr__ = __str__


def _preprocess_value(value: t.Any) -> t.Any:
    if isinstance(value, PulpEntityContext):
        return value.pulp_href
    if isinstance(value, t.Mapping):
        return {k: _preprocess_value(v) for k, v in value.items()}
    if isinstance(value, (list, tuple)):
        return [_preprocess_value(item) for item in value]
    return value


def preprocess_payload(payload: EntityDefinition) -> EntityDefinition:
    if isinstance(payload, PreprocessedEntityDefinition):
        return payload

    return PreprocessedEntityDefinition(
        {key: _preprocess_value(value) for key, value in payload.items() if value is not None}
    )


_REGISTERED_API_QUIRKS: t.List[t.Tuple[PluginRequirement, t.Callable[[OpenAPI], None]]] = []


def api_quirk(
    req: PluginRequirement,
) -> t.Callable[[t.Callable[[OpenAPI], None]], None]:
    """
    A function decorator to allow manipulating API specs based on the availability of plugins.

    Parameters:
        req: The plugin specifier to determine when the quirk should be applied.

    Examples:
        ```
        @api_quirk(PluginRequirement("catdog", specifier="<1.5.2"))
        def patch_barking_filter_type(api: OpenAPI) -> None:
            # fixup api.api_spec here
        ```
    """

    def _decorator(patch: t.Callable[[OpenAPI], None]) -> None:
        _REGISTERED_API_QUIRKS.append((req, patch))

    return _decorator


@api_quirk(PluginRequirement("core", specifier="<3.20.0"))
def patch_ordering_filters(api: OpenAPI) -> None:
    for method, path in api.operations.values():
        operation = api.api_spec["paths"][path][method]
        if method == "get" and "parameters" in operation:
            for parameter in operation["parameters"]:
                if (
                    parameter["name"] == "ordering"
                    and parameter["in"] == "query"
                    and "schema" in parameter
                    and parameter["schema"]["type"] == "string"
                ):
                    parameter["schema"] = {"type": "array", "items": {"type": "string"}}
                    parameter["explode"] = False
                    parameter["style"] = "form"


@api_quirk(PluginRequirement("core", specifier="<3.22.0"))
def patch_field_select_filters(api: OpenAPI) -> None:
    for method, path in api.operations.values():
        operation = api.api_spec["paths"][path][method]
        if method == "get" and "parameters" in operation:
            for parameter in operation["parameters"]:
                if (
                    parameter["name"] in ["fields", "exclude_fields"]
                    and parameter["in"] == "query"
                    and "schema" in parameter
                    and parameter["schema"]["type"] == "string"
                ):
                    parameter["schema"] = {"type": "array", "items": {"type": "string"}}


@api_quirk(PluginRequirement("core", specifier="<99.99.0"))
def patch_content_in_query_filters(api: OpenAPI) -> None:
    # https://github.com/pulp/pulpcore/issues/3634
    for operation_id, (method, path) in api.operations.items():
        if (
            operation_id == "repository_versions_list"
            or (
                operation_id.startswith("repositories_") and operation_id.endswith("_versions_list")
            )
            or (operation_id.startswith("publications_") and operation_id.endswith("_list"))
        ):
            operation = api.api_spec["paths"][path][method]
            for parameter in operation["parameters"]:
                if (
                    parameter["name"] == "content__in"
                    and parameter["in"] == "query"
                    and "schema" in parameter
                    and parameter["schema"]["type"] == "string"
                ):
                    parameter["schema"] = {"type": "array", "items": {"type": "string"}}


@api_quirk(PluginRequirement("core", specifier=">=3.23,<3.30.0"))
def patch_upstream_pulp_replicate_request_body(api: OpenAPI) -> None:
    operation = api.api_spec["paths"]["{upstream_pulp_href}replicate/"]["post"]
    operation.pop("requestBody", None)


class PulpContext:
    """
    Abstract class for the global PulpContext object.
    It is an abstraction layer for api access and output handling.

    Parameters:
        api_root: The base url (excluding "api/v3/") to the servers api.
        api_kwargs: Extra arguments to pass to the wrapped `OpenAPI` object.
        background_tasks: Whether to wait for tasks. If `True`, all tasks triggered will
            immediately raise `PulpNoWait`.
        timeout: Limit of time to wait for unfinished tasks.
        domain: Name of the domain to interact with.
        fake_mode: In fake mode, no modifying calls will be performed.
            Where possible, instead of failing, the requested result will be faked.
            This implies `dry_run=True` on the `api_kwargs`.
        verify_ssl: A boolean or a path to the CA bundle.
    """

    def echo(self, message: str, nl: bool = True, err: bool = False) -> None:
        """
        Abstract function that will be called to emit warnings and task progress.

        Warning:
            This function does nothing until implemented by a subclass.
        """

    def prompt(self, text: str, hide_input: bool = False) -> str:
        """
        Abstract function that will be called to ask for a password interactively.

        Note:
            If a password is provided non-interactively, this function need not be implemented.
            Doing so is deprecated.
        """
        raise NotImplementedError("PulpContext is an abstract class.")

    def __init__(
        self,
        api_root: str,
        api_kwargs: t.Dict[str, t.Any],
        background_tasks: bool = False,
        timeout: t.Union[int, datetime.timedelta] = 300,
        domain: str = "default",
        fake_mode: bool = False,
        verify_ssl: t.Optional[t.Union[bool, str]] = None,
        verify: t.Optional[t.Union[bool, str]] = None,  # Deprecated
    ) -> None:
        self._api: t.Optional[OpenAPI] = None
        self._api_root: str = api_root
        self._api_kwargs = api_kwargs
        self.verify_ssl = verify_ssl
        if self.verify_ssl is None:
            # Regrets, regrets...
            self.verify_ssl = (
                verify if verify is not None else self._api_kwargs.pop("validate_certs", True)
            )
        if self.verify_ssl is True:
            # If this is "only" true and we have the PULP_CA_BUNDLE variable set, use it.
            self.verify_ssl = os.environ.get("PULP_CA_BUNDLE", True)
        self._needed_plugins: t.List[PluginRequirement] = [
            PluginRequirement("core", specifier=">=3.11.0")
        ]
        self.pulp_domain: str = domain

        self.background_tasks: bool = background_tasks
        if isinstance(timeout, int):
            timeout = datetime.timedelta(seconds=timeout)
        self.timeout: datetime.timedelta = timeout
        self.fake_mode: bool = fake_mode
        if self.fake_mode:
            self._api_kwargs["dry_run"] = True

    @classmethod
    def from_config_files(
        cls,
        profile: t.Optional[str] = None,
        config_locations: t.Optional[t.List[str]] = None,
        overrides: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> "t.Self":
        """
        Create a `PulpContext` object from config files.

        Note:
            This feature needs Python >=3.11 to work for now, because we don't want to add another
            dependency to pulp-glue.

        Parameters:
            profile: Select a different profile from the config.
            config_locations: If provided these config files will be merged (last on wins) instead
                of the default locations.
            overrides: Values with higer priority than that from the config files.
        Returns:
            A configured `PulpContext` object.
        """
        if config_locations is None:
            config_locations = [
                "/etc/pulp/cli.toml",
                (os.environ.get("XDG_CONFIG_HOME") or os.environ["HOME"] + "/.config")
                + "/pulp/cli.toml",
            ]
        configs: t.Dict[str, t.Dict[str, t.Any]] = {}
        for location in config_locations:
            try:
                with open(location, "rb") as fp:
                    new_configs = tomllib.load(fp)
            except (FileNotFoundError, PermissionError):
                pass
            else:
                # level 1 merge
                for key in new_configs:
                    if key in configs:
                        configs[key].update(new_configs[key])
                    else:
                        configs[key] = new_configs[key]
        profile_key: str = "cli" if profile is None else "cli-" + profile
        config = configs[profile_key]
        if overrides is not None:
            config.update(overrides)
        return cls.from_config(config)

    @classmethod
    def from_config(cls, config: t.Dict[str, t.Any]) -> "t.Self":
        """
        Create a `PulpContext` object from a config dictionary.

        Parameters:
            config: dictionary of configuration values.
        Returns:
            A configured `PulpContext` object.
        """
        api_kwargs: t.Dict[str, t.Any] = {
            "base_url": config["base_url"],
        }
        if "username" in config:
            api_kwargs["auth_provider"] = BasicAuthProvider(config["username"], config["password"])
        if "headers" in config:
            api_kwargs["headers"] = dict(
                (header.split(":", maxsplit=1) for header in config["headers"])
            )
        for key in ["cert", "key", "user_agent", "cid", "dry_run"]:
            if key in config:
                api_kwargs[key] = config[key]

        return cls(
            api_root=config.get("api_root", "/pulp/"),
            domain=config.get("domain", "default"),
            verify_ssl=config.get("verify_ssl", True),
            api_kwargs=api_kwargs,
        )

    def _patch_api_spec(self) -> None:
        # A place for last minute fixes to the api_spec.
        # WARNING: Operations are already indexed at this point.
        assert self._api is not None
        for req, patch in _REGISTERED_API_QUIRKS:
            if self.has_plugin(req):
                patch(self._api)

    @property
    def domain_enabled(self) -> bool:
        return t.cast(bool, self.api.api_spec.get("info", {}).get("x-pulp-domain-enabled", False))

    @property
    def api_path(self) -> str:
        if self.domain_enabled:
            return self._api_root + self.pulp_domain + "/api/v3/"
        return self._api_root + "api/v3/"

    @property
    def api(self) -> OpenAPI:
        """
        The lazy evaluated `OpenAPI` object contained in this context.

        This is only needed for low level interactions with the openapi spec.
        All calls to the API should be performed via `call`.
        """
        if self._api is None:
            username = self._api_kwargs.pop("username", None)
            password = self._api_kwargs.pop("password", None)
            if username:
                # Deprecated for 'auth'.
                if not password:
                    password = self.prompt("password", hide_input=True)
                self._api_kwargs["auth_provider"] = BasicAuthProvider(username, password)
                warnings.warn(
                    "Using 'username' and 'password' with 'PulpContext' is deprecated. "
                    "Use an auth provider with the 'auth_provider' argument instead.",
                    DeprecationWarning,
                )
            try:
                self._api = OpenAPI(
                    doc_path=f"{self._api_root}api/v3/docs/api.json",
                    verify_ssl=self.verify_ssl,
                    **self._api_kwargs,
                )
            except OpenAPIError as e:
                raise PulpException(str(e))
            # Rerun scheduled version checks
            for plugin_requirement in self._needed_plugins:
                self.needs_plugin(plugin_requirement)
            self._patch_api_spec()
        return self._api

    @property
    def component_versions(self) -> t.Dict[str, str]:
        result: t.Dict[str, str] = self.api.api_spec.get("info", {}).get("x-pulp-app-versions", {})
        return result

    def call(
        self,
        operation_id: str,
        non_blocking: bool = False,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        body: t.Optional[EntityDefinition] = None,
        validate_body: bool = True,
    ) -> t.Any:
        """
        Perform an API call for operation_id.
        Wait for triggered tasks to finish if not background.
        Returns the operation result, or the finished task.

        Parameters:
            operation_id: The operation ID in the openapi v3 spec to be called.
            non_blocking: returns unfinished tasks if `True`.
            parameters: Arguments that are to be sent as headers, querystrings or part of the URI.
            body: Body payload for POST, PUT, PATCH calls.
            validate_body: Indicate whether the body should be validated.

        Returns:
            The body of the response, or the task or task group if one was issued.

        Raises:
            PulpNoWait: in case the context has `background_tasks` set or a task (group) timed out.
            NotImplementedFake: if an unsafe call was attempted in `fake_mode`
            PulpHTTPError: for unhandeld REST API errors
            PulpException: for all unhandeld openapi and http connection exceptions
        """
        if parameters is None:
            parameters = {}
        if self.domain_enabled:
            # Validation will fail if path doesn't need domain parameter
            if "pulp_domain" in self.api.param_spec(operation_id, "path", required=True):
                parameters["pulp_domain"] = self.pulp_domain
        parameters = preprocess_payload(parameters)
        if body is not None:
            body = preprocess_payload(body)
        try:
            result = self.api.call(
                operation_id,
                parameters=parameters,
                body=body,
                validate_body=validate_body,
            )
        except UnsafeCallError:
            if self.fake_mode:
                raise NotImplementedFake(f"Operation {operation_id} was attempted in fake mode.")
            else:
                raise
        # Asynchronous tasks seem to be reported by a dict containing only one key "task"
        if isinstance(result, dict) and ["task"] == list(result.keys()):
            task_href = result["task"]
            result = self.api.call("tasks_read", parameters={"task_href": task_href})
            self.echo(
                _("Started background task {task_href}").format(task_href=task_href), err=True
            )
            if not non_blocking:
                result = self.wait_for_task(result)
        if isinstance(result, dict) and ["task_group"] == list(result.keys()):
            task_group_href = result["task_group"]
            result = self.api.call(
                "task_groups_read", parameters={"task_group_href": task_group_href}
            )
            self.echo(
                _("Started background task group {task_group_href}").format(
                    task_group_href=task_group_href
                ),
                err=True,
            )
            if not non_blocking:
                result = self.wait_for_task_group(result)
        return result

    @staticmethod
    def _task_finished(task: EntityDefinition, expect_cancel: bool = False) -> bool:
        task_href = task["pulp_href"]

        if task["state"] == "completed":
            if expect_cancel:
                raise PulpException(
                    _("The task {task_href} meant to be canceled, completed instead.")
                )
            return True
        elif task["state"] == "failed":
            raise PulpException(
                _("Task {task_href} failed: '{description}'").format(
                    task_href=task_href,
                    description=task["error"].get("description") or task["error"].get("reason"),
                )
            )
        elif task["state"] == "canceled":
            if expect_cancel:
                return True
            raise PulpException(_("Task {task_href} canceled").format(task_href=task_href))
        elif task["state"] in ["waiting", "running", "canceling"]:
            return False
        else:
            raise NotImplementedError(_("Unknown task state: {state}").format(state=task["state"]))

    def wait_for_task(self, task: EntityDefinition, expect_cancel: bool = False) -> t.Any:
        """
        Wait for a task to finish and return the finished task object.

        Parameters:
            task: A task object to monitor.
            expect_cancel: Swaps the raising condition for completed and canceled tasks.

        Raises:
            PulpNoWait: on timeout or if the context has `background_tasks` set.
            PulpException: on ctrl-c, if task failed or was canceled.
        """
        deadline = datetime.datetime.now() + self.timeout if self.timeout else None

        if self.background_tasks:
            raise PulpNoWait(_("Not waiting for task because --background was specified."))
        task_href = task["pulp_href"]
        try:
            while not self._task_finished(task, expect_cancel=expect_cancel):
                if deadline and datetime.datetime.now() > deadline:
                    raise PulpNoWait(
                        _("Waiting for task {task_href} timed out.").format(
                            task_href=task["pulp_href"]
                        )
                    )
                time.sleep(1)
                self.echo(".", nl=False, err=True)
                task = self.api.call("tasks_read", parameters={"task_href": task["pulp_href"]})
            self.echo("Done.", err=True)
            return task
        except KeyboardInterrupt:
            raise PulpNoWait(_("Task {task_href} sent to background.").format(task_href=task_href))

    def _task_group_finished(self, task_group: EntityDefinition) -> bool:
        if task_group["waiting"] + task_group["running"] + task_group["canceling"] > 0:
            return False
        if task_group["failed"] + task_group["canceled"] > 0:
            errors = []
            for task in task_group["tasks"]:
                if task["state"] in ["failed", "canceled"]:
                    task = self.api.call("tasks_read", parameters={"task_href": task["pulp_href"]})
                    errors.append(task["error"].get("description") or task["error"].get("reason"))
            raise PulpException(
                _("Task group {task_group_href} has failed/canceled tasks: '{errors}'").format(
                    task_group_href=task_group["pulp_href"],
                    errors="\n".join(errors),
                )
            )
        return True

    def wait_for_task_group(self, task_group: EntityDefinition) -> t.Any:
        """
        Wait for a task group to finish and return the finished task object.

        Parameters:
            task_group: A task group object all of which's tasks to monitor.

        Raises:
            PulpNoWait: on timeout or if the context has `background_tasks` set.
            PulpException: on ctrl-c, if task failed or was canceled.
        """
        deadline = datetime.datetime.now() + self.timeout if self.timeout else None

        if self.background_tasks:
            raise PulpNoWait("Not waiting for task group because --background was specified.")
        try:
            while not self._task_group_finished(task_group):
                if deadline and datetime.datetime.now() > deadline:
                    raise PulpNoWait(
                        _("Waiting for task group {task_group_href} timed out.").format(
                            task_group_href=task_group["pulp_href"]
                        )
                    )
                time.sleep(1)
                self.echo(".", nl=False, err=True)
                task_group = self.api.call(
                    "task_groups_read", parameters={"task_group_href": task_group["pulp_href"]}
                )
        except KeyboardInterrupt:
            raise PulpNoWait(
                _("Task group {task_group_href} sent to background.").format(
                    task_group_href=task_group["pulp_href"]
                )
            )

    def has_plugin(
        self,
        plugin_requirement: PluginRequirement,
    ) -> bool:
        version: t.Optional[str] = self.component_versions.get(plugin_requirement.name)
        return version in plugin_requirement

    def needs_plugin(
        self,
        plugin_requirement: PluginRequirement,
    ) -> None:
        if self._api is not None:
            if not self.has_plugin(plugin_requirement):
                component = f"{plugin_requirement.name}{plugin_requirement.specifier}"
                feature = plugin_requirement.feature or _("this command")
                if plugin_requirement.inverted:
                    msg = _(
                        "The server provides the pulp component '{component}',"
                        " which prevents the use of {feature}."
                        " See 'pulp status' for installed components."
                    )
                else:
                    msg = _(
                        "The server does not provide the pulp component '{component}',"
                        " which is needed to use {feature}."
                        " See 'pulp status' for installed components."
                    )
                raise PulpException(msg.format(component=component, feature=feature))
        else:
            # Schedule for later checking
            self._needed_plugins.append(plugin_requirement)


class PulpViewSetContext:
    """
    Base class to interact with a generic viewset.

    Parameters:
        pulp_ctx: The server context to attach this viewset context to.
    """

    # Subclasses should provide appropriate values here
    ID_PREFIX: t.ClassVar[str]
    """Common prefix for the operations of this entity."""
    NEEDS_PLUGINS: t.ClassVar[t.List[PluginRequirement]] = []
    """List of plugin requirements to operate such an entity on the server."""

    def __init__(self, pulp_ctx: PulpContext) -> None:
        self.pulp_ctx: PulpContext = pulp_ctx

        # Add requirements to the lazy evaluated list
        for plugin_requirement in self.NEEDS_PLUGINS:
            self.pulp_ctx.needs_plugin(plugin_requirement)

    def call(
        self,
        operation: str,
        non_blocking: bool = False,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        body: t.Optional[EntityDefinition] = None,
        validate_body: bool = True,
    ) -> t.Any:
        """
        Perform an API call for operation.
        Wait for triggered tasks to finish if not background.
        Returns the operation result, or the finished task.

        Parameters:
            operation: The operation to be performed on the entity. Usually the openapi
                operation_id is constructed by concatenating with the `ID_PREFIX`.
            non_blocking: returns unfinished tasks if `True`.
            parameters: Arguments that are to be sent as headers, querystrings or part of the URI.
            body: Body payload for POST, PUT, PATCH calls.
            validate_body: Indicate whether the body should be validated.

        Returns:
            The body of the response, or the task or task group if one was issued.

        Raises:
            PulpNoWait: in case the context has `background_tasks` set or a task (group) timed out.
        """
        operation_id: str = (
            getattr(self, operation.upper() + "_ID", None) or self.ID_PREFIX + "_" + operation
        )
        return self.pulp_ctx.call(
            operation_id,
            non_blocking=non_blocking,
            parameters=parameters,
            body=body,
            validate_body=validate_body,
        )


class PulpEntityContext(PulpViewSetContext):
    """
    Base class for entity specific contexts.
    This class provides the basic CRUD commands and ties its instances to the global
    PulpContext for api access.
    It typically corresponds to a NamedModelViewset.
    Mostly specification is achieved by defining / extending the class attributes below.

    Parameters:
        pulp_ctx: The server context to attach this entity context to.
        pulp_href: Specifying this is equivalent to assinging to `pulp_href` later.
        entity: Specifying this is equivalent to assinging to `entity` later.
    """

    # Subclasses should provide appropriate values here
    ENTITY: t.ClassVar[str] = _("entity")
    """Translatable name of the entity."""
    ENTITIES: t.ClassVar[str] = _("entities")
    """Translatable plural of `ENTITY`."""
    HREF: t.ClassVar[str]
    """Name of the href parameter in the url patterns."""
    NULLABLES: t.ClassVar[t.Set[str]] = set()
    """Set of fields that can be cleared by sending 'null'."""
    CAPABILITIES: t.ClassVar[t.Dict[str, t.List[PluginRequirement]]] = {}
    """
    List of capabilities this entity provides.

    Subclasses can specify version dependent capabilities here

    Example:
        ```
        CAPABILITIES = {
            "feature1": [
                PluginRequirement("file"),
                PluginRequirement("core", specifier=">=3.7.0")
            ]
        }
        ```
    """
    HREF_PATTERN: str
    """Regular expression with capture groups for 'plugin' and 'resource_type' to match URIs."""

    # Hidden values for the lazy entity lookup
    _entity: t.Optional[EntityDefinition]
    _entity_lookup: EntityDefinition

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        """
        Extra scope used in parameters for create and list calls.

        Subclasses for nested entities can define the parameters for there parent scope here.
        """
        return {}

    @property
    def entity(self) -> EntityDefinition:
        """
        Entity property that will perform a lazy lookup once it is accessed.
        You can specify lookup parameters by assigning a dictionary to it,
        or assign an href to the `pulp_href` property.
        To reset to having no attached entity you can assign `None`.

        !!!note:
            `A # type: ignore[assignment]` comment is needed due to a mypy limitation.
            https://github.com/python/mypy/issues/3004

        Assigning to it will reset the lazy lookup behaviour.
        """
        if self._entity is None:
            if not self._entity_lookup:
                raise PulpException(
                    _("A {entity} must be specified for this command.").format(entity=self.ENTITY)
                )
            if self._entity_lookup.get("pulp_href"):
                self._entity = self.show(self._entity_lookup["pulp_href"])
            else:
                self._entity = self.find(**self._entity_lookup)
            self._entity_lookup = {}
        return self._entity

    @entity.setter
    def entity(self, value: t.Optional[EntityDefinition]) -> None:
        # Setting this property will always (lazily) retrigger retrieving the entity.
        # If set multiple times in a row without reading, the criteria will be added.
        if value is None:
            self._entity_lookup = {}
        else:
            self._entity_lookup.update(value)
            self._entity_lookup.pop("pulp_href", None)
        self._entity = None

    @property
    def pulp_href(self) -> str:
        """
        Property to represent the href of the attached entity.
        Assigning to it will reset the lazy lookup behaviour.
        """
        return str(self.entity["pulp_href"])

    @pulp_href.setter
    def pulp_href(self, value: str) -> None:
        if not href_regex.fullmatch(value):
            raise PulpException(
                _("'{value}' is not a valid HREF value for a {context_id}").format(
                    value=value, context_id=self.ENTITY
                )
            )

        # Setting this property will always (lazily) retrigger retrieving the entity.
        self._entity_lookup = {"pulp_href": value}
        self._entity = None

    @property
    def tangible(self) -> bool:
        """Indicate whether an entity is available or specified by search parameters."""
        return bool(self._entity) or bool(self._entity_lookup)

    def __init__(
        self,
        pulp_ctx: PulpContext,
        pulp_href: t.Optional[str] = None,
        entity: t.Optional[EntityDefinition] = None,
    ) -> None:
        assert pulp_href is None or entity is None

        super().__init__(pulp_ctx)
        self.meta: t.Dict[str, str] = {}

        self._entity = None
        self._entity_lookup = entity or {}
        if pulp_href is not None:
            self.pulp_href = pulp_href

    @classmethod
    def _preprocess_value(cls, key: str, value: t.Any) -> t.Any:
        if key in cls.NULLABLES and value == "":
            return None
        return _preprocess_value(value)

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        """
        Filter to prepare the body for a create or update call.

        This function can be subclassed by specific Entity contexts to fix data depending on plugin
        versions.

        Parameters:
            body: The payload representing the entity to create or the fields to update on it.
            partial: Should be set if `body` may only represent part of the entity.

        Returns:
            The body ready to be passed to `call`.
        """
        if isinstance(body, PreprocessedEntityDefinition):
            return body

        return PreprocessedEntityDefinition(
            {
                key: self._preprocess_value(key, value)
                for key, value in body.items()
                if value is not None
            }
        )

    def list_iterator(
        self,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        offset: int = 0,
        batch_size: int = BATCH_SIZE,
        stats: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Iterator[t.Any]:
        """
        List entities from this context in a batched iterator.

        Parameters:
            parameters: Search or sorting criteria.
            offset: Number of entities to skip.
            batch_size: Size of the batches to fetch.
                Maximally BATCH_SIZE will be used.
            stats: If provided, a dictionary that will be filled with metadata:
                count: Number of entities reported by the server to match the criteria.

        Returns:
            Iterator of entities matching the search conditions.
        """
        payload: t.Dict[str, t.Any] = parameters.copy() if parameters else {}
        payload.update(self.scope)
        payload["offset"] = offset
        payload["limit"] = min(batch_size, BATCH_SIZE)
        while True:
            response: t.Mapping[str, t.Any] = self.call("list", parameters=payload)
            if stats is not None:
                stats["count"] = response["count"]
            payload["offset"] += len(response["results"])
            yield from response["results"]
            if response["next"] is None:
                break

    def list(self, limit: int, offset: int, parameters: t.Dict[str, t.Any]) -> t.List[t.Any]:
        """
        List entities by the type of this context.

        Parameters:
            limit: Maximal number of entities to return
                Use 0 to loop until all entries are retrieved.
            offset: Number of entities to skip at the front of the list.
            parameters: Additional search or sorting criteria.

        Returns:
            List of entities matching the conditions.
        """
        entities = []
        stats: t.Dict[str, t.Any] = {}
        try:
            if limit > 0:
                iterator = self.list_iterator(
                    parameters=parameters, offset=offset, batch_size=limit, stats=stats
                )
                for i in range(limit):
                    entities.append(next(iterator))
            else:
                entities.extend(
                    self.list_iterator(parameters=parameters, offset=offset, stats=stats)
                )
        except StopIteration:
            pass
        else:
            self.pulp_ctx.echo(
                _("Not all {count} entries were shown.").format(count=stats["count"]), err=True
            )
        return entities

    def find(self, **kwargs: t.Any) -> t.Any:
        """
        Find an entity based on search parameters.

        Note: It is preferred to use the `entity` property instead of calling `find` directly.

        Parameters:
            kwargs: The search parameters.

        Returns:
            The entity if one was found uniquely.

        Raises:
            PulpEntityNotFound: if no entity satisfies the search parameters.
            PulpException: if multiple entities satisfy the search parameters.
        """
        payload: t.Dict[str, t.Any] = kwargs.copy()
        payload.update(self.scope)
        payload["offset"] = 0
        payload["limit"] = 1
        result: t.Mapping[str, t.Any] = self.call("list", parameters=payload)
        if result["count"] == 0:
            raise PulpEntityNotFound(
                _("Could not find {entity} with {kwargs}.").format(
                    entity=self.ENTITY, kwargs=kwargs
                )
            )
        if result["count"] > 1:
            raise PulpException(
                _("Multiple {entities} found with {kwargs}.").format(
                    entities=self.ENTITIES, kwargs=kwargs
                )
            )
        return result["results"][0]

    def show(self, href: t.Optional[str] = None) -> t.Any:
        """
        Retrieve and return the full record of an entity from the server.

        Parameters:
            href: href of the entity to fetch. If not specified, the entity represented by `entity`
                will be used.

        Returns:
            The full record of the entity as reported by the server.

        Raises:
            PulpException: if no href was specified and the lazy lookup failed.
        """
        return self.call("read", parameters={self.HREF: href or self.pulp_href})

    def create(
        self,
        body: EntityDefinition,
        parameters: t.Optional[t.Mapping[str, t.Any]] = None,
        non_blocking: bool = False,
    ) -> t.Any:
        """
        Create an entity.

        Parameters:
            body: Fields off the new entity and the values they should be set to.
            parameters: Additional parameters for the call (usually not needed).
            non_blocking: Whether the result of the operation should be awaited on.

        Returns:
            The created entity, or the record of the create task if non_blocking.
        """
        _parameters = self.scope
        if parameters:
            _parameters.update(parameters)
        if body is not None:
            body = self.preprocess_entity(body, partial=False)
        if self.pulp_ctx.fake_mode:
            body["pulp_href"] = "<FAKE ENTITY>"
            self._entity = body
            self._entity_lookup = {}
            return self._entity

        result = self.call(
            "create",
            parameters=_parameters,
            body=body,
            non_blocking=non_blocking,
        )
        if result["pulp_href"].startswith(self.pulp_ctx.api_path + "tasks/"):
            if not non_blocking:
                try:
                    if getattr(self, "HREF_PATTERN", None):
                        self.pulp_href = next(
                            (
                                h
                                for h in result["created_resources"]
                                if re.match(
                                    re.escape(self.pulp_ctx.api_path) + self.HREF_PATTERN, h
                                )
                            )
                        )
                    else:
                        self.pulp_href = result["created_resources"][0]  # YOLO
                    result = self.entity
                except (KeyError, StopIteration):
                    raise PulpException(_("No suitable resource got created."))
            else:
                self.entity = None
        else:
            self._entity = result
            self._entity_lookup = {}

        return result

    def update(
        self,
        body: t.Optional[EntityDefinition] = None,
        parameters: t.Optional[t.Mapping[str, t.Any]] = None,
        non_blocking: bool = False,
    ) -> t.Any:
        """
        Update the entity.

        Parameters:
            body: Fields and the values they should be changed to.
            parameters: Additional parameters for the call (usually not needed).
            non_blocking: Whether the result of the operation should be awaited on.

        Returns:
            The updated entity, or the record of the update task if non_blocking.
        """
        _parameters = {self.HREF: self.pulp_href}
        if parameters:
            _parameters.update(parameters)
        if body is not None:
            body = self.preprocess_entity(body, partial=False)
        if self.pulp_ctx.fake_mode:
            assert self._entity is not None
            if body is not None:
                self._entity.update(body)
            return self._entity

        result = self.call(
            "partial_update",
            parameters=_parameters,
            body=body,
            non_blocking=non_blocking,
        )
        if result["pulp_href"].startswith(self.pulp_ctx.api_path + "tasks/"):
            self.pulp_href = self.pulp_href  # reenable lazy lookup
            if not non_blocking:
                result = self.entity  # reload from server
        else:
            self._entity = result
            self._entity_lookup = {}

        return result

    def delete(self, non_blocking: bool = False) -> t.Any:
        """
        Delete the entity.

        Parameters:
            non_blocking: Whether the result of the operation should be awaited on.

        Returns:
            The record of the delete task.
        """
        if self.pulp_ctx.fake_mode:
            self.entity = None
            return None

        result = self.call(
            "delete", parameters={self.HREF: self.pulp_href}, non_blocking=non_blocking
        )
        self.entity = None
        return result

    def set_label(self, key: str, value: str, non_blocking: bool = False) -> t.Any:
        """
        Set a label.

        Parameters:
            key: Name of the label.
            value: Value of the label.
            non_blocking: Whether the result of the operation should be awaited on.
                This no longer relevant for pulpcore>=3.34, because the synchronous api is used.
        """
        if self.pulp_ctx.fake_mode:
            self.entity  # perform lookup
            assert self._entity is not None
            self._entity["pulp_labels"][key] = value
            return None
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.34.0")):
            try:
                return self.call(
                    "set_label",
                    parameters={self.HREF: self.pulp_href},
                    body={"key": key, "value": value},
                )
            except PulpHTTPError as e:
                if e.status_code != 403:
                    raise
                # Workaround for broken access policies: Try the old mechanism.
        labels = self.entity["pulp_labels"]
        labels[key] = value
        return self.update(body={"pulp_labels": labels}, non_blocking=non_blocking)

    def unset_label(self, key: str, non_blocking: bool = False) -> t.Any:
        """
        Unset a label.

        Parameters:
            key: Name of the label.
            non_blocking: Whether the result of the operation should be awaited on.
                This no longer relevant for pulpcore>=3.34, because the synchronous api is used.
        """
        if self.pulp_ctx.fake_mode:
            self.entity  # perform lookup
            assert self._entity is not None
            self._entity["pulp_labels"].pop(key)
            return None
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.34.0")):
            try:
                return self.call(
                    "unset_label", parameters={self.HREF: self.pulp_href}, body={"key": key}
                )
            except PulpHTTPError as e:
                if e.status_code != 403:
                    raise
                # Workaround for broken access policies: Try the old mechanism.
        labels = self.entity["pulp_labels"]
        try:
            labels.pop(key)
        except KeyError:
            raise PulpException(_("Could not find label with key '{key}'.").format(key=key))
        return self.update(body={"pulp_labels": labels}, non_blocking=non_blocking)

    def show_label(self, key: str) -> t.Optional[str]:
        """
        Show value of a label.

        Parameters:
            key: Name of the label.

        Returns:
            Value of the label or None.

        Raises:
            PulpException: if the label was not set.
        """
        labels: t.Dict[str, t.Optional[str]] = self.entity["pulp_labels"]
        try:
            return labels[key]
        except KeyError:
            raise PulpException(_("Could not find label with key '{key}'.").format(key=key))

    def my_permissions(self) -> t.Any:
        self.needs_capability("roles")
        return self.call("my_permissions", parameters={self.HREF: self.pulp_href})

    def list_roles(self) -> t.Any:
        self.needs_capability("roles")
        return self.call("list_roles", parameters={self.HREF: self.pulp_href})

    def add_role(self, role: str, users: t.List[str], groups: t.List[str]) -> t.Any:
        self.needs_capability("roles")
        return self.call(
            "add_role",
            parameters={self.HREF: self.pulp_href},
            body={"role": role, "users": users, "groups": groups},
        )

    def remove_role(self, role: str, users: t.List[str], groups: t.List[str]) -> t.Any:
        self.needs_capability("roles")
        return self.call(
            "remove_role",
            parameters={self.HREF: self.pulp_href},
            body={"role": role, "users": users, "groups": groups},
        )

    def converge(
        self,
        desired_attributes: t.Optional[t.Dict[str, t.Any]],
        defaults: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Tuple[bool, t.Optional[EntityDefinition], t.Optional[EntityDefinition]]:
        """
        Converge an entity to have a set of desired attributes.

        This will look for the entity, and depending on what it found and what should be, create,
        delete or update the entity.

        Parameters:
            desired_attributes: Dictionary of attributes the entity should have. `None` if the
                entity is supposed to be absent.
            defaults: Optional dict with default and extra values to be used when creating a new
                entity.

        Returns:
            Tuple of (changed, before, after)
        """
        try:
            entity = self.entity
        except PulpEntityNotFound:
            entity = None

        if desired_attributes is None:
            if entity is not None:
                self.delete()
                return True, entity, None
        else:
            if entity is None:
                desired_entity: t.Dict[str, t.Any] = {}
                if defaults:
                    desired_entity.update(defaults)
                desired_entity.update(desired_attributes)
                desired_entity.update(self._entity_lookup)
                desired_entity.pop("pulp_href", None)
                return True, None, self.create(desired_entity)
            else:
                update_attributes = {}
                for k, v in self.preprocess_entity(desired_attributes, partial=True).items():
                    if entity.get(k) != v:
                        update_attributes[k] = v
                if update_attributes:
                    return (
                        True,
                        entity,
                        self.update(PreprocessedEntityDefinition(update_attributes)),
                    )
        return False, entity, entity

    def capable(self, capability: str) -> bool:
        """
        Report on a capability based on the presence of all needed server plugins.

        Parameters:
            capability: Name of a capability.

        Returns:
            Whether the capability is provided for this context.
        """
        return capability in self.CAPABILITIES and all(
            (self.pulp_ctx.has_plugin(pr) for pr in self.CAPABILITIES[capability])
        )

    def needs_capability(self, capability: str) -> None:
        """
        Translates a capability in calls to `needs_plugin` via `CAPABILITIES`.

        Parameters:
            capability: Name of a capability.
        """
        if capability in self.CAPABILITIES:
            for plugin_requirement in self.CAPABILITIES[capability]:
                self.pulp_ctx.needs_plugin(plugin_requirement)
        else:
            raise PulpException(
                _("Capability '{capability}' needed on '{entity}' for this command.").format(
                    capability=capability, entity=self.ENTITY
                )
            )


class PulpRemoteContext(PulpEntityContext):
    """
    Base class for remote contexts.
    """

    ENTITY = _("remote")
    ENTITIES = _("remotes")
    ID_PREFIX = "remotes"
    HREF_PATTERN = r"remotes/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    NULLABLES = {
        "ca_cert",
        "client_cert",
        "client_key",
        "username",
        "password",
        "proxy_url",
        "proxy_username",
        "proxy_password",
        "download_concurrency",
        "max_retries",
        "total_timeout",
        "connect_timeout",
        "sock_connect_timeout",
        "sock_read_timeout",
        "rate_limit",
    }
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpRemoteContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls


class PulpPublicationContext(PulpEntityContext):
    """Base class for publication contexts."""

    ENTITY = _("publication")
    ENTITIES = _("publications")
    ID_PREFIX = "publications"
    HREF_PATTERN = r"publications/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpPublicationContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls

    def list(self, limit: int, offset: int, parameters: t.Dict[str, t.Any]) -> t.List[t.Any]:
        if parameters.get("repository") is not None:
            self.pulp_ctx.needs_plugin(
                PluginRequirement("core", specifier=">=3.20.0", feature=_("repository filter"))
            )
        return super().list(limit, offset, parameters)


class PulpDistributionContext(PulpEntityContext):
    """Base class for distribution contexts."""

    ENTITY = _("distribution")
    ENTITIES = _("distributions")
    ID_PREFIX = "distributions"
    HREF_PATTERN = r"distributions/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    NULLABLES = {"content_guard", "publication", "remote", "repository", "repository_version"}
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpDistributionContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls


class PulpRepositoryVersionContext(PulpEntityContext):
    """
    Base class for repository version contexts.

    Parameters:
        pulp_ctx: The server context to attach this entity to.
        repository_ctx: Context of the repository this context should be scoped to.
        pulp_href: Specifying this is equivalent to assinging to `pulp_href` later.
    """

    ENTITY = _("repository version")
    ENTITIES = _("repository versions")
    ID_PREFIX = "repository_versions"
    repository_ctx: "PulpRepositoryContext"

    def __init__(
        self,
        pulp_ctx: PulpContext,
        repository_ctx: "PulpRepositoryContext",
        pulp_href: t.Optional[str] = None,
    ) -> None:
        super().__init__(pulp_ctx, pulp_href=pulp_href)
        self.repository_ctx = repository_ctx

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        if self.ID_PREFIX == "repository_versions":
            return {}
        else:
            return {self.repository_ctx.HREF: self.repository_ctx.pulp_href}

    @property
    def entity(self) -> EntityDefinition:
        if (
            self._entity is None
            and "number" in self._entity_lookup.keys()
            and self._entity_lookup.get("number") is None
        ):
            self.pulp_href = self.repository_ctx.entity["latest_version_href"]
        return super().entity

    @entity.setter
    def entity(self, value: t.Optional[EntityDefinition]) -> None:
        # ignore needed due to a bug regarding overriding property setter in mypy
        PulpEntityContext.entity.fset(self, value)  # type: ignore[attr-defined]

    def repair(self) -> t.Any:
        """
        Trigger a repair task for this repository version.

        Returns:
            The record of the repair task.
        """
        return self.call("repair", parameters={self.HREF: self.pulp_href}, body={})


class PulpRepositoryContext(PulpEntityContext):
    """Base class for repository contexts."""

    ENTITY = _("repository")
    ENTITIES = _("repositories")
    HREF_PATTERN = r"repositories/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    ID_PREFIX = "repositories"
    VERSION_CONTEXT: t.ClassVar[t.Type[PulpRepositoryVersionContext]] = PulpRepositoryVersionContext
    NULLABLES = {"description", "remote", "retain_repo_versions"}
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpRepositoryContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls

    def get_version_context(
        self,
        number: t.Optional[int] = None,
    ) -> PulpRepositoryVersionContext:
        """
        Return a repository version context of the proper type scoped for this repository.

        Parameters:
            number: Version number or `-1` for the latest version.

        Returns:
            Repository version context attached to the same pulp_ctx and scoped to the repository.
        """
        if number is not None:
            if number >= 0:
                version_href = self.entity["versions_href"] + f"{number}/"
            elif number == -1:
                version_href = self.entity["latest_version_href"]
            else:
                PulpException(_("Invalid version number ({number}).").format(number=number))
        else:
            version_href = None
        return self.VERSION_CONTEXT(
            pulp_ctx=self.pulp_ctx, repository_ctx=self, pulp_href=version_href
        )

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if "retain_repo_versions" in body:
            self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.13.0"))
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.13.0,<3.15.0")):
            # "retain_repo_versions" has been named "retained_versions" until pulpcore 3.15
            # https://github.com/pulp/pulpcore/pull/1472
            if "retain_repo_versions" in body:
                body["retained_versions"] = body.pop("retain_repo_versions")
        return body

    def sync(self, body: t.Optional[EntityDefinition] = None) -> t.Any:
        """
        Trigger a sync task for this repository.

        Parameters:
            body: Any additional options specific to the repository type used to perform this sync.

        Returns:
            Record of the sync task.
        """
        return self.call("sync", parameters={self.HREF: self.pulp_href}, body=body or {})

    def modify(
        self,
        add_content: t.Optional[t.List[str]] = None,
        remove_content: t.Optional[t.List[str]] = None,
        base_version: t.Optional[t.Union[str, PulpRepositoryVersionContext]] = None,
    ) -> t.Any:
        """
        Add to or remove content from this repository.

        Parameters:
            add_content: List of content hrefs to add.
            remove_content: List of content hrefs to remove.
            base_version: Href to a repository version relative to whose content the changes are to
                be interpreted.

        Returns:
            Record of the modify task.
        """
        body: t.Dict[str, t.Any] = {}
        if add_content is not None:
            body["add_content_units"] = add_content
        if remove_content is not None:
            body["remove_content_units"] = remove_content
        if base_version is not None:
            body["base_version"] = base_version
        if self.pulp_ctx.fake_mode:
            return {"state": "completed"}  # Fake a task
        return self.call("modify", parameters={self.HREF: self.pulp_href}, body=body)


class PulpGenericRepositoryContext(PulpRepositoryContext):
    """Generic repository context class to separate specific general functionality."""

    def reclaim(
        self,
        repo_hrefs: t.List[t.Union[str, PulpRepositoryContext]],
        repo_versions_keeplist: t.Optional[
            t.List[t.Union[str, PulpRepositoryVersionContext]]
        ] = None,
    ) -> t.Any:
        """
        Reclaim disk space for a list of repositories.

        Parameters:
            repo_hrefs: List of repository hrefs to reclaim.
            repo_versions_keeplist: List of repository version hrefs to keep unaffected.

        Returns:
            Record of the reclaim space task.
        """
        self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.19.0"))
        body: t.Dict[str, t.Any] = {}
        body["repo_hrefs"] = repo_hrefs
        if repo_versions_keeplist:
            body["repo_versions_keeplist"] = repo_versions_keeplist
        return self.call("reclaim_space_reclaim", body=body)


class PulpContentContext(PulpEntityContext):
    """Base class for content contexts."""

    ENTITY = _("content")
    ENTITIES = _("content")
    ID_PREFIX = "content"
    HREF_PATTERN = r"content/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpContentContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls

    def __init__(
        self,
        pulp_ctx: PulpContext,
        pulp_href: t.Optional[str] = None,
        entity: t.Optional[EntityDefinition] = None,
        repository_ctx: t.Optional[PulpRepositoryContext] = None,
    ):
        super().__init__(pulp_ctx, pulp_href=pulp_href, entity=entity)
        self.repository_ctx = repository_ctx

    def list(self, limit: int, offset: int, parameters: t.Dict[str, t.Any]) -> t.List[t.Any]:
        if self.repository_ctx is not None:
            parameters = parameters.copy()
            parameters["repository_version"] = self.repository_ctx.entity["latest_version_href"]
        return super().list(limit, offset, parameters)

    def find(self, **kwargs: t.Any) -> t.Any:
        if self.repository_ctx is not None:
            kwargs["repository_version"] = self.repository_ctx.entity["latest_version_href"]
        return super().find(**kwargs)

    def create(
        self,
        body: EntityDefinition,
        parameters: t.Optional[t.Mapping[str, t.Any]] = None,
        non_blocking: bool = False,
    ) -> t.Any:
        with ExitStack() as cleanup:
            body = body.copy()
            file = body.pop("file", None)
            chunk_size: t.Optional[int] = body.pop("chunk_size", None)
            if file:
                if isinstance(file, str):
                    file = open(file, "rb")
                    cleanup.enter_context(file)
                size = os.path.getsize(file.name)
                if not self.pulp_ctx.fake_mode:  # Skip the uploading part in fake_mode
                    if chunk_size is None or chunk_size > size:
                        body["file"] = file
                    elif self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.20.0")):
                        self.needs_capability("upload")
                        from pulp_glue.core.context import PulpUploadContext

                        upload_href = PulpUploadContext(self.pulp_ctx).upload_file(file, chunk_size)
                        body["upload"] = upload_href
                    else:
                        from pulp_glue.core.context import PulpArtifactContext

                        artifact_href = PulpArtifactContext(self.pulp_ctx).upload(file, chunk_size)
                        body["artifact"] = artifact_href
            if self.repository_ctx is not None:
                body["repository"] = self.repository_ctx
            return super().create(body=body, parameters=parameters, non_blocking=non_blocking)

    def delete(self, non_blocking: bool = False) -> None:
        assert self.repository_ctx is not None
        self.repository_ctx.modify(remove_content=[self.pulp_href])

    def upload(
        self,
        file: t.IO[bytes],
        chunk_size: int,
        repository: t.Optional[PulpRepositoryContext],
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Create a content unit by uploading a file.

        This function is deprecated. The create call can handle the upload logic transparently.

        Parameters:
            file: A file like object that supports `os.path.getsize`.
            chunk_size: Size of the chunks to upload independently.
            repository: Repository context to add the newly created content to.
            kwargs: Extra args specific to the content type, passed to the create call.

        Returns:
            The result of the create task.
        """
        self.needs_capability("upload")
        size = os.path.getsize(file.name)
        body: t.Dict[str, t.Any] = {**kwargs}
        if not self.pulp_ctx.fake_mode:  # Skip the uploading part in fake_mode
            if chunk_size > size:
                body["file"] = file
            elif self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.20.0")):
                from pulp_glue.core.context import PulpUploadContext

                upload_href = PulpUploadContext(self.pulp_ctx).upload_file(file, chunk_size)
                body["upload"] = upload_href
            else:
                from pulp_glue.core.context import PulpArtifactContext

                artifact_href = PulpArtifactContext(self.pulp_ctx).upload(file, chunk_size)
                body["artifact"] = artifact_href
            if repository:
                body["repository"] = repository
        return self.create(body=body)


class PulpACSContext(PulpEntityContext):
    """Base class for ACS contexts."""

    ENTITY = _("ACS")
    ENTITIES = _("ACSes")
    HREF_PATTERN = r"acs/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    ID_PREFIX = "acs"
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpACSContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls

    def refresh(self, href: t.Optional[str] = None) -> t.Any:
        return self.call("refresh", parameters={self.HREF: href or self.pulp_href})


class PulpContentGuardContext(PulpEntityContext):
    """Base class for content guard contexts."""

    ENTITY = "content guard"
    ENTITIES = "content guards"
    ID_PREFIX = "contentguards"
    HREF_PATTERN = r"contentguards/(?P<plugin>[\w\-_]+)/(?P<resource_type>[\w\-_]+)/"
    NULLABLES = {"description"}
    TYPE_REGISTRY: t.Final[t.Dict[str, t.Type["PulpContentGuardContext"]]] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "PLUGIN") and hasattr(cls, "RESOURCE_TYPE"):
            cls.TYPE_REGISTRY[f"{cls.PLUGIN}:{cls.RESOURCE_TYPE}"] = cls


EntityFieldDefinition = t.Union[None, str, PulpEntityContext]



import typing as t

from pulp_glue.common.context import (
    EntityDefinition,
    PluginRequirement,
    PulpContentContext,
    PulpDistributionContext,
    PulpEntityContext,
    PulpRemoteContext,
    PulpRepositoryContext,
    PulpRepositoryVersionContext,
)
from pulp_glue.common.i18n import get_translation

translation = get_translation(__package__)
_ = translation.gettext


class PulpContainerBlobContext(PulpContentContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "blob"
    ENTITY = _("container blob")
    ENTITIES = _("container blobs")
    HREF = "container_blob_href"
    ID_PREFIX = "content_container_blobs"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]


class PulpContainerManifestContext(PulpContentContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "manifest"
    ENTITY = _("container manifest")
    ENTITIES = _("container manifests")
    HREF = "container_manifest_href"
    ID_PREFIX = "content_container_manifests"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]


class PulpContainerTagContext(PulpContentContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "tag"
    ENTITY = _("container tag")
    ENTITIES = _("container tags")
    HREF = "container_tag_href"
    ID_PREFIX = "content_container_tags"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]

    def find(self, **kwargs: t.Any) -> t.Any:
        if "digest" in kwargs and isinstance(kwargs["digest"], str):
            kwargs["digest"] = [kwargs["digest"]]
        return super().find(**kwargs)


class PulpContainerNamespaceContext(PulpEntityContext):
    ENTITY = _("container namespace")
    ENTITIES = _("container namespaces")
    HREF = "container_container_namespace_href"
    ID_PREFIX = "pulp_container_namespaces"
    HREF_PATTERN = r"(?P<plugin>pulp_container)/(?P<resource_type>namespaces)/"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]
    CAPABILITIES = {"roles": [PluginRequirement("container", specifier=">=2.11.0")]}


class PulpContainerDistributionContext(PulpDistributionContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "container"
    ENTITY = _("container distribution")
    ENTITIES = _("container distributions")
    HREF = "container_container_distribution_href"
    ID_PREFIX = "distributions_container_container"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]
    CAPABILITIES = {"roles": [PluginRequirement("container", specifier=">=2.11.0")]}

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        version = body.pop("version", None)
        if version is not None:
            repository_href = body.pop("repository")
            body["repository_version"] = f"{repository_href}versions/{version}/"
        return body


class PulpContainerRemoteContext(PulpRemoteContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "container"
    ENTITY = _("container remote")
    ENTITIES = _("container remotes")
    HREF = "container_container_remote_href"
    ID_PREFIX = "remotes_container_container"
    NULLABLES = PulpRemoteContext.NULLABLES | {"include_tags", "exclude_tags"}
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]
    CAPABILITIES = {"roles": [PluginRequirement("container", specifier=">=2.11.0")]}


class PulpContainerRepositoryVersionContext(PulpRepositoryVersionContext):
    HREF = "container_container_repository_version_href"
    ID_PREFIX = "repositories_container_container_versions"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]


class PulpContainerPushRepositoryVersionContext(PulpRepositoryVersionContext):
    HREF = "container_container_push_repository_version_href"
    ID_PREFIX = "repositories_container_container_push_versions"
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]


class PulpContainerBaseRepositoryContext(PulpRepositoryContext):
    NEEDS_PLUGINS = [PluginRequirement("container", specifier=">=2.3.0")]

    def tag(self, tag: str, digest: str) -> t.Any:
        self.needs_capability("tag")
        return self.call(
            "tag",
            parameters={self.HREF: self.pulp_href},
            body={"tag": tag, "digest": digest},
        )

    def untag(self, tag: str) -> t.Any:
        self.needs_capability("tag")
        return self.call(
            "untag",
            parameters={self.HREF: self.pulp_href},
            body={"tag": tag},
        )


class PulpContainerRepositoryContext(PulpContainerBaseRepositoryContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "container"
    HREF = "container_container_repository_href"
    ID_PREFIX = "repositories_container_container"
    ENTITY = _("container repository")
    ENTITIES = _("container repositories")
    VERSION_CONTEXT = PulpContainerRepositoryVersionContext
    HREF_PATTERN = r"repositories/(?P<plugin>container)/(?P<resource_type>container)/"
    CAPABILITIES = {
        "sync": [PluginRequirement("container")],
        "pulpexport": [PluginRequirement("container", specifier=">=2.8.0")],
        "tag": [PluginRequirement("container", specifier=">=2.3.0")],
        "roles": [PluginRequirement("container", specifier=">=2.11.0")],
    }

    def modify(
        self,
        add_content: t.Optional[t.List[str]] = None,
        remove_content: t.Optional[t.List[str]] = None,
        base_version: t.Optional[t.Union[str, PulpRepositoryVersionContext]] = None,
    ) -> t.Any:
        if remove_content:
            self.call(
                "remove",
                parameters={self.HREF: self.pulp_href},
                body={"content_units": remove_content},
            )
        if add_content:
            self.call(
                "add",
                parameters={self.HREF: self.pulp_href},
                body={"content_units": add_content},
            )

    def copy_tag(self, source_href: str, tags: t.Optional[t.List[str]]) -> t.Any:
        body = {"source_repository_version": source_href, "names": tags}
        return self.call("copy_tags", parameters={self.HREF: self.pulp_href}, body=body)

    def copy_manifest(
        self,
        source_href: str,
        digests: t.Optional[t.List[str]],
        media_types: t.Optional[t.List[str]],
    ) -> t.Any:
        body = {
            "source_repository_version": source_href,
            "digests": digests,
            "media_types": media_types,
        }
        return self.call("copy_manifests", parameters={self.HREF: self.pulp_href}, body=body)


class PulpContainerPushRepositoryContext(PulpContainerBaseRepositoryContext):
    PLUGIN = "container"
    RESOURCE_TYPE = "push"
    HREF = "container_container_push_repository_href"
    ID_PREFIX = "repositories_container_container_push"
    ENTITY = _("push container repository")
    ENTITIES = _("push container repositories")
    VERSION_CONTEXT = PulpContainerPushRepositoryVersionContext
    HREF_PATTERN = r"repositories/(?P<plugin>container)/(?P<resource_type>container-push)/"
    CAPABILITIES = {
        "tag": [PluginRequirement("container", specifier=">=2.3.0")],
        "roles": [PluginRequirement("container", specifier=">=2.11.0")],
        "remove": [PluginRequirement("container", specifier=">=2.4.0")],
    }

    def remove_image(self, digest: str) -> t.Any:
        self.needs_capability("remove")
        body = {"digest": digest}
        return self.call("remove_image", parameters={self.HREF: self.pulp_href}, body=body)



import gettext
import sys
from functools import lru_cache

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files


# Need to call lru_cache() before using it as a decorator for python 3.7 compatibility
@lru_cache(maxsize=None)
def get_translation(name: str) -> gettext.NullTranslations:
    """
    Return a translations object for a certain import path.

    Parameters:
        name: An import path.

    Returns:
        A `gettext` translations object containing all the `gettext` variations.

    Examples:
        Import the usual suspects like this:
        ```
        from pulp_glue.common.i18n import get_translation

        translation = get_translation(__package__)
        _ = translation.gettext
        ```
    """
    name_parts = name.split(".")
    while name_parts:
        try:
            localedir = files(".".join(name_parts)) / "locale"
            return gettext.translation("messages", localedir=str(localedir), fallback=True)
        except TypeError:
            name_parts.pop()
    raise TypeError(f"No package found for {name}.")



import datetime
import hashlib
import os
import sys
import typing as t

from pulp_glue.common.context import (
    EntityDefinition,
    PluginRequirement,
    PulpContentGuardContext,
    PulpContext,
    PulpEntityContext,
    PulpViewSetContext,
    preprocess_payload,
)
from pulp_glue.common.exceptions import PulpException
from pulp_glue.common.i18n import get_translation

translation = get_translation(__package__)
_ = translation.gettext


class PulpAccessPolicyContext(PulpEntityContext):
    ENTITY = _("access policy")
    ENTITIES = _("access policies")
    HREF = "access_policy_href"
    ID_PREFIX = "access_policies"

    def reset(self) -> t.Any:
        self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.17.0"))
        return self.call("reset", parameters={self.HREF: self.pulp_href})

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            if "creation_hooks" in body:
                body["permissions_assignment"] = body.pop("creation_hooks")
        return body


class PulpArtifactContext(PulpEntityContext):
    ENTITY = _("artifact")
    ENTITIES = _("artifacts")
    HREF = "artifact_href"
    ID_PREFIX = "artifacts"

    def upload(
        self, file: t.IO[bytes], chunk_size: int = 1000000, sha256: t.Optional[str] = None
    ) -> t.Any:
        size = os.path.getsize(file.name)

        sha256_hasher = hashlib.sha256()
        for chunk in iter(lambda: file.read(chunk_size), b""):
            sha256_hasher.update(chunk)
        sha256_digest = sha256_hasher.hexdigest()
        file.seek(0)

        if sha256 is not None and sha256_digest != sha256:
            raise PulpException(_("File digest does not match."))

        # check whether the file exists before uploading
        result = self.list(limit=1, offset=0, parameters={"sha256": sha256_digest})
        if len(result) > 0:
            self.pulp_ctx.echo(_("Artifact already exists."), err=True)
            self.pulp_href = result[0]["pulp_href"]
            return result[0]["pulp_href"]

        self.pulp_ctx.echo(_("Uploading file {filename}").format(filename=file.name), err=True)

        if self.pulp_ctx.fake_mode:
            self._entity = {"pulp_href": "<FAKE_ENTITY>", "sha256": sha256, "size": size}
            self._entity_lookup = {}
            return self._entity["pulp_href"]
        if chunk_size > size:
            # if chunk_size is bigger than the file size, just upload it directly
            artifact: t.Dict[str, t.Any] = self.create({"sha256": sha256_digest, "file": file})
            self.pulp_href = artifact["pulp_href"]
            return artifact["pulp_href"]

        upload_ctx = PulpUploadContext(self.pulp_ctx)
        upload_ctx.upload_file(file, chunk_size)

        self.pulp_ctx.echo(_("Creating artifact."), err=True)
        try:
            task = upload_ctx.commit(sha256_digest)
        except Exception as e:
            upload_ctx.delete()
            raise e
        self.pulp_href = task["created_resources"][0]
        return task["created_resources"][0]


class PulpDomainContext(PulpEntityContext):
    ENTITY = _("Pulp domain")
    ENTITIES = _("Pulp domains")
    HREF = "domain_href"
    ID_PREFIX = "domains"
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.23.0")]


class PulpExporterContext(PulpEntityContext):
    ENTITY = _("Pulp exporter")
    ENTITIES = _("Pulp exporters")
    HREF = "pulp_exporter_href"
    ID_PREFIX = "exporters_core_pulp"


class PulpExportContext(PulpEntityContext):
    ENTITY = _("Pulp export")
    ENTITIES = _("Pulp exports")
    HREF = "pulp_pulp_export_href"
    ID_PREFIX = "exporters_core_pulp_exports"
    exporter: EntityDefinition

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        return {PulpExporterContext.HREF: self.exporter["pulp_href"]}


class PulpGroupContext(PulpEntityContext):
    ENTITY = _("user group")
    ENTITIES = _("user groups")
    # Handled by a workaround
    # HREF = "group_href"
    ID_PREFIX = "groups"
    CAPABILITIES = {"roles": [PluginRequirement("core", specifier=">=3.17.0")]}

    @property
    def HREF(self) -> str:  # type:ignore
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            return "auth_group_href"
        return "group_href"

    def add_user(self, user: "PulpUserContext") -> None:
        pass

    def remove_user(self, user: "PulpUserContext") -> None:
        user_pk = user.pulp_href.split("/")[-2]
        group_user_ctx = PulpGroupUserContext(self.pulp_ctx, self)
        # This is a bad workaround, because the api is terrible.
        group_user_ctx._entity = {
            "pulp_href": f"{group_user_ctx.group_ctx.pulp_href}users/{user_pk}/"
        }
        group_user_ctx.delete()


class PulpGroupPermissionContext(PulpEntityContext):
    ENTITY = _("group permission")
    ENTITIES = _("group permissions")
    NEEDS_PLUGINS = [PluginRequirement("core", specifier="<3.20.0", feature=_("group permissions"))]
    group_ctx: PulpGroupContext

    def __init__(self, pulp_ctx: PulpContext, group_ctx: PulpGroupContext) -> None:
        super().__init__(pulp_ctx)
        self.group_ctx = group_ctx

    def call(
        self,
        operation: str,
        non_blocking: bool = False,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        body: t.Optional[t.Dict[str, t.Any]] = None,
        validate_body: bool = False,
    ) -> t.Any:
        # Workaroud because the openapi spec for GroupPermissions has always been broken.
        #
        # This will probably not be fixed upstream, and GroupPermissions are removed from pulpcore.
        # So we just skip linting here.
        return super().call(
            operation,
            non_blocking=non_blocking,
            parameters=parameters,
            body=body,
            validate_body=validate_body,
        )

    def find(self, **kwargs: t.Any) -> t.Any:
        # Workaroud the missing ability to filter.
        # # TODO fix upstream and adjust to guard for the proper version
        # # https://pulp.plan.io/issues/8241
        # if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.99.dev")):
        #     # Workaround not needed anymore
        #     return super().find(**kwargs)
        search_result = self.list(limit=sys.maxsize, offset=0, parameters={})
        for key, value in kwargs.items():
            search_result = [res for res in search_result if res[key] == value]
        if len(search_result) != 1:
            raise PulpException(
                _("Could not find {entity} with {kwargs}.").format(
                    entity=self.ENTITY, kwargs=kwargs
                )
            )
        return search_result[0]

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        return {self.group_ctx.HREF: self.group_ctx.pulp_href}


class PulpGroupModelPermissionContext(PulpGroupPermissionContext):
    ENTITY = _("group model permission")
    ENTITIES = _("group model permissions")
    # Handled by a workaround
    # HREF = "groups_model_permission_href"
    ID_PREFIX = "groups_model_permissions"

    @property
    def HREF(self) -> str:  # type:ignore
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            return "auth_groups_model_permission_href"
        return "groups_model_permission_href"


class PulpGroupObjectPermissionContext(PulpGroupPermissionContext):
    ENTITY = _("group object permission")
    ENTITIES = _("group object permissions")
    # Handled by a workaround
    # HREF = "groups_object_permission_href"
    ID_PREFIX = "groups_object_permissions"

    @property
    def HREF(self) -> str:  # type:ignore
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            return "auth_groups_object_permission_href"
        return "groups_object_permission_href"


class PulpGroupRoleContext(PulpEntityContext):
    ENTITY = _("group role")
    ENTITIES = _("group roles")
    HREF = "groups_group_role_href"
    ID_PREFIX = "groups_roles"
    NULLABLES = {"content_object"}
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.17.0", feature=_("group roles"))]
    group_ctx: PulpGroupContext

    def __init__(self, pulp_ctx: PulpContext, group_ctx: PulpGroupContext) -> None:
        super().__init__(pulp_ctx)
        self.group_ctx = group_ctx

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        return {self.group_ctx.HREF: self.group_ctx.pulp_href}


class PulpGroupUserContext(PulpEntityContext):
    ENTITY = _("group user")
    ENTITIES = _("group users")
    # Handled by a workaround
    # HREF = "groups_user_href"
    ID_PREFIX = "groups_users"
    group_ctx: PulpGroupContext

    @property
    def HREF(self) -> str:  # type:ignore
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            return "auth_groups_user_href"
        return "groups_user_href"

    def __init__(self, pulp_ctx: PulpContext, group_ctx: PulpGroupContext) -> None:
        super().__init__(pulp_ctx)
        self.group_ctx = group_ctx

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        return {self.group_ctx.HREF: self.group_ctx.pulp_href}


class PulpImporterContext(PulpEntityContext):
    ENTITY = _("Pulp importer")
    ENTITIES = _("Pulp importers")
    HREF = "pulp_importer_href"
    ID_PREFIX = "importers_core_pulp"


class PulpOrphanContext(PulpViewSetContext):
    ID_PREFIX = "orphans_cleanup"

    def cleanup(self, body: t.Optional[t.Dict[str, t.Any]] = None) -> t.Any:
        if body is not None:
            body = preprocess_payload(body)
            if "orphan_protection_time" in body:
                self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.15.0"))
        else:
            body = {}
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.14.0")):
            result = self.call("cleanup", body=body)
        else:
            if body:
                self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.14.0"))
            if not self.pulp_ctx.fake_mode:
                result = self.pulp_ctx.call("orphans_delete")
            else:
                # Do we need something better?
                result = {}
        return result


class PulpCompositeContentGuardContext(PulpContentGuardContext):
    PLUGIN = "core"
    RESOURCE_TYPE = "composite"
    ENTITY = "composite content guard"
    ENTITIES = "composite content guards"
    HREF = "composite_content_guard_href"
    ID_PREFIX = "contentguards_core_composite"
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.43.0")]


class PulpContentRedirectContentGuardContext(PulpContentGuardContext):
    PLUGIN = "core"
    RESOURCE_TYPE = "content_redirect"
    ENTITY = "content redirect content guard"
    ENTITIES = "content redirect content guards"
    HREF = "content_redirect_content_guard_href"
    ID_PREFIX = "contentguards_core_content_redirect"
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.18.0")]


class PulpHeaderContentGuardContext(PulpContentGuardContext):
    PLUGIN = "core"
    RESOURCE_TYPE = "header"
    ENTITY = "header content guard"
    ENTITIES = "header content guards"
    HREF = "header_content_guard_href"
    ID_PREFIX = "contentguards_core_header"
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.39.0")]


class PulpRbacContentGuardContext(PulpContentGuardContext):
    PLUGIN = "core"
    RESOURCE_TYPE = "rbac"
    ENTITY = "RBAC content guard"
    ENTITIES = "RBAC content guards"
    HREF = "r_b_a_c_content_guard_href"
    ID_PREFIX = "contentguards_core_rbac"
    DOWNLOAD_ROLE: t.ClassVar[str] = "core.rbaccontentguard_downloader"
    CAPABILITIES = {"roles": [PluginRequirement("core", specifier=">=3.17.0")]}
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.15.0")]

    def assign(
        self,
        href: t.Optional[str] = None,
        users: t.Optional[t.List[str]] = None,
        groups: t.Optional[t.List[str]] = None,
    ) -> t.Any:
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            body: EntityDefinition = {"users": users, "groups": groups}
            body["role"] = self.DOWNLOAD_ROLE
            return self.call("add_role", parameters={self.HREF: href or self.pulp_href}, body=body)
        else:
            body = {"usernames": users, "groupnames": groups}
            return self.call(
                "assign_permission", parameters={self.HREF: href or self.pulp_href}, body=body
            )

    def remove(
        self,
        href: t.Optional[str] = None,
        users: t.Optional[t.List[str]] = None,
        groups: t.Optional[t.List[str]] = None,
    ) -> t.Any:
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.17.0")):
            body: EntityDefinition = {"users": users, "groups": groups}
            body["role"] = self.DOWNLOAD_ROLE
            return self.call(
                "remove_role", parameters={self.HREF: href or self.pulp_href}, body=body
            )
        else:
            body = {"usernames": users, "groupnames": groups}
            return self.call(
                "remove_permission", parameters={self.HREF: href or self.pulp_href}, body=body
            )


class PulpRoleContext(PulpEntityContext):
    ENTITY = _("role")
    ENTITIES = _("roles")
    HREF = "role_href"
    ID_PREFIX = "roles"
    NULLABLES = {"description"}
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.17.0")]


class PulpSigningServiceContext(PulpEntityContext):
    ENTITY = _("signing service")
    ENTITIES = _("signing services")
    HREF = "signing_service_href"
    ID_PREFIX = "signing_services"
    HREF_PATTERN = r"signing-services/"


class PulpTaskContext(PulpEntityContext):
    ENTITY = _("task")
    ENTITIES = _("tasks")
    HREF = "task_href"
    ID_PREFIX = "tasks"
    CAPABILITIES = {"roles": [PluginRequirement("core", specifier=">=3.17.0")]}

    resource_context: t.Optional[PulpEntityContext] = None

    def list(self, limit: int, offset: int, parameters: t.Dict[str, t.Any]) -> t.List[t.Any]:
        if (
            parameters.get("logging_cid") is not None
            or parameters.get("logging_cid__contains") is not None
        ):
            self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.14.0"))
        if not self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.22.0")):
            parameters = parameters.copy()
            reserved_resources = parameters.pop("reserved_resources", None)
            exclusive_resources = parameters.pop("exclusive_resources", None)
            shared_resources = parameters.pop("shared_resources", None)
            if (
                parameters.pop("reserved_resources__in", None)
                or parameters.pop("exclusive_resources__in", None)
                or parameters.pop("shared_resources__in", None)
            ):
                self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.22.0"))
            reserved_resources_record = []
            if reserved_resources:
                reserved_resources_record.append(reserved_resources)
            if exclusive_resources:
                reserved_resources_record.append(exclusive_resources)
            if shared_resources:
                reserved_resources_record.append("shared:" + shared_resources)
            if len(reserved_resources_record) > 1:
                self.pulp_ctx.needs_plugin(
                    PluginRequirement(
                        "core",
                        specifier=">=3.22.0",
                        feature=_("specify multiple reserved resources"),
                    ),
                )
            parameters["reserved_resources_record"] = reserved_resources_record

        return super().list(limit=limit, offset=offset, parameters=parameters)

    def cancel(self, task_href: t.Optional[str] = None, background: bool = False) -> t.Any:
        task_href = task_href or self.pulp_href
        task = self.call(
            "cancel",
            parameters={self.HREF: task_href},
            body={"state": "canceled"},
        )
        if not background:
            self.pulp_ctx.echo(_("Waiting to cancel task {href}").format(href=task_href), err=True)
            task = self.pulp_ctx.wait_for_task(task, expect_cancel=True)
            self.pulp_ctx.echo(_("Done."), err=True)
        return task

    def profile_artifact_urls(self) -> t.Dict[str, str]:
        self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.57.0"))
        result = self.call(
            "profile_artifacts",
            parameters={self.HREF: self.pulp_href},
        )["urls"]
        assert isinstance(result, dict)
        return result

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        if self.resource_context:
            if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.22.0")):
                return {"reserved_resources": self.resource_context.pulp_href}
            else:
                return {"reserved_resources_record": [self.resource_context.pulp_href]}
        else:
            return {}

    def purge(
        self,
        finished_before: t.Optional[datetime.datetime],
        states: t.Optional[t.List[str]],
    ) -> t.Any:
        self.pulp_ctx.needs_plugin(PluginRequirement("core", specifier=">=3.17.0"))
        body: t.Dict[str, t.Any] = {}
        if finished_before:
            body["finished_before"] = finished_before
        if states:
            body["states"] = states
        return self.call(
            "purge",
            body=body,
        )

    def summary(self) -> t.Dict[str, int]:
        task_states = ["waiting", "skipped", "running", "completed", "failed", "canceled"]
        if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.14.0")):
            task_states.append("canceling")
        result = {}
        for state in task_states:
            payload = {"limit": 1, "state": state}
            result[state] = self.call("list", parameters=payload)["count"]
        return result


class PulpTaskGroupContext(PulpEntityContext):
    ENTITY = _("task group")
    ENTITIES = _("task groups")
    HREF = "task_group_href"
    ID_PREFIX = "task_groups"


class PulpUploadContext(PulpEntityContext):
    ENTITY = _("upload")
    ENTITIES = _("uploads")
    HREF = "upload_href"
    ID_PREFIX = "uploads"

    def upload_chunk(
        self,
        chunk: bytes,
        size: int,
        start: int,
        non_blocking: bool = False,
    ) -> t.Any:
        end: int = start + len(chunk) - 1
        parameters = {self.HREF: self.pulp_href, "Content-Range": f"bytes {start}-{end}/{size}"}
        return self.call(
            "update",
            parameters=parameters,
            body={"sha256": hashlib.sha256(chunk).hexdigest(), "file": chunk},
            non_blocking=non_blocking,
        )

    def commit(self, sha256: str) -> t.Any:
        return self.call(
            "commit",
            parameters={self.HREF: self.pulp_href},
            body={"sha256": sha256},
        )

    def upload_file(self, file: t.IO[bytes], chunk_size: int = 1000000) -> t.Any:
        """Upload a file and return the uncommitted upload_href."""
        start = 0
        size = os.path.getsize(file.name)
        upload_href = self.create(body={"size": size})["pulp_href"]
        try:
            self.pulp_href = upload_href
            while start < size:
                chunk = file.read(chunk_size)
                self.upload_chunk(
                    chunk=chunk,
                    size=size,
                    start=start,
                )
                start += chunk_size
                self.pulp_ctx.echo(".", nl=False, err=True)
        except Exception as e:
            self.delete(upload_href)
            raise e
        self.pulp_ctx.echo(_("Upload complete."), err=True)
        return upload_href


class PulpUserContext(PulpEntityContext):
    ENTITY = _("user")
    ENTITIES = _("users")
    HREF = "auth_user_href"
    ID_PREFIX = "users"
    NULLABLES = {"password"}


class PulpUserRoleContext(PulpEntityContext):
    ENTITY = _("user role")
    ENTITIES = _("user roles")
    HREF = "auth_users_user_role_href"
    ID_PREFIX = "users_roles"
    NULLABLES = {"content_object"}
    user_ctx: PulpUserContext

    def __init__(self, pulp_ctx: PulpContext, user_ctx: PulpUserContext) -> None:
        super().__init__(pulp_ctx)
        self.user_ctx = user_ctx

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        return {self.user_ctx.HREF: self.user_ctx.pulp_href}


class PulpWorkerContext(PulpEntityContext):
    ENTITY = _("worker")
    ENTITIES = _("workers")
    HREF = "worker_href"
    ID_PREFIX = "workers"
    HREF_PATTERN = r"workers/"


class PulpUpstreamPulpContext(PulpEntityContext):
    ENTITY = _("upstream pulp")
    ENTITIES = _("upstream pulps")
    HREF = "upstream_pulp_href"
    ID_PREFIX = "upstream_pulps"
    HREF_PATTERN = r"upstream-pulps/"
    NEEDS_PLUGINS = [PluginRequirement("core", specifier=">=3.23.0")]

    def find(self, **kwargs: t.Any) -> t.Any:
        # Workaroud the missing ability to filter.
        # # TODO fix upstream and adjust to guard for the proper version
        # # https://github.com/pulp/pulpcore/issues/4110
        # if self.pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.99.dev")):
        #     # Workaround not needed anymore
        #     return super().find(**kwargs)
        search_result = self.list(limit=sys.maxsize, offset=0, parameters={})
        for key, value in kwargs.items():
            search_result = [res for res in search_result if res[key] == value]
        if len(search_result) != 1:
            raise PulpException(
                _("Could not find {entity} with {kwargs}.").format(
                    entity=self.ENTITY, kwargs=kwargs
                )
            )
        return search_result[0]

    def replicate(self) -> t.Any:
        return self.call("replicate", parameters={self.HREF: self.pulp_href})



import typing as t

import pytest

from pulp_glue.common.authentication import OAuth2ClientCredentialsAuth

pytestmark = pytest.mark.glue


def test_sending_no_scope_when_empty(monkeypatch: pytest.MonkeyPatch) -> None:

    class OAuth2MockResponse:
        def raise_for_status(self) -> None:
            return None

        def json(self) -> t.Dict[str, t.Any]:
            return {"expires_in": 1, "access_token": "aaa"}

    def _requests_post_mocked(
        url: str, data: t.Dict[str, t.Any], **kwargs: t.Any
    ) -> OAuth2MockResponse:
        assert "scope" not in data
        return OAuth2MockResponse()

    monkeypatch.setattr("requests.post", _requests_post_mocked)

    OAuth2ClientCredentialsAuth(token_url="", client_id="", client_secret="")._retrieve_token()

    OAuth2ClientCredentialsAuth(
        token_url="", client_id="", client_secret="", scopes=[]
    )._retrieve_token()



import json
import logging
import typing as t

import pytest

from pulp_glue.common.openapi import OpenAPI

TEST_SCHEMA = json.dumps(
    {
        "openapi": "3.0.3",
        "paths": {"test/": {"get": {"operationId": "test_id", "responses": {200: {}}}}},
        "components": {"schemas": {}},
    }
).encode()


class MockRequest:
    headers: t.Dict[str, str] = {}
    body: t.Dict[str, t.Any] = {}


class MockResponse:
    status_code = 200
    headers: t.Dict[str, str] = {}
    text = "{}"
    content: t.Dict[str, t.Any] = {}

    def raise_for_status(self) -> None:
        pass


class MockSession:
    def prepare_request(self, *args: t.Any, **kwargs: t.Any) -> MockRequest:
        return MockRequest()

    def send(self, request: MockRequest) -> MockResponse:
        return MockResponse()


@pytest.fixture
def openapi(monkeypatch: pytest.MonkeyPatch) -> OpenAPI:
    monkeypatch.setattr(OpenAPI, "load_api", lambda self, refresh_cache: TEST_SCHEMA)
    openapi = OpenAPI("base_url", "doc_path")
    openapi._parse_api(TEST_SCHEMA)
    monkeypatch.setattr(openapi, "_session", MockSession())
    return openapi


def test_openapi_logs_nothing_from_info(openapi: OpenAPI, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.INFO)
    openapi.call("test_id")
    assert caplog.record_tuples == []


def test_openapi_logs_operation_info_to_debug(
    openapi: OpenAPI, caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.DEBUG)
    openapi.call("test_id")
    assert caplog.record_tuples == [
        ("pulp_glue.openapi", logging.DEBUG + 3, "test_id : get test/"),
        ("pulp_glue.openapi", logging.DEBUG + 2, ""),
        ("pulp_glue.openapi", logging.DEBUG + 1, "{}"),
        ("pulp_glue.openapi", logging.DEBUG + 3, "Response: 200"),
        ("pulp_glue.openapi", logging.DEBUG + 1, "{}"),
    ]



from copy import deepcopy

import pytest

from pulp_glue.common.context import _REGISTERED_API_QUIRKS, PulpContext

pytestmark = pytest.mark.glue


@pytest.mark.live
def test_api_quirks_idempotent(pulp_ctx: PulpContext) -> None:
    """
    Test, that the applied api quirks can be applied twice without failing.

    This can let us hope they will not fail once the api is fixed upstream.
    """
    assert {
        "patch_content_in_query_filters",
        "patch_field_select_filters",
        "patch_file_acs_refresh_request_body",
        "patch_upstream_pulp_replicate_request_body",
        "patch_python_remote_includes_excludes",
        "patch_ordering_filters",
    } <= {quirk[1].__name__ for quirk in _REGISTERED_API_QUIRKS}

    patched_once_api = deepcopy(pulp_ctx.api.api_spec)
    # Patch a second time
    assert pulp_ctx._api is not None
    pulp_ctx._patch_api_spec()
    assert pulp_ctx.api.api_spec == patched_once_api
